# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for loading GGD grids from IMAS grid_ggd IDS structures."""

from typing import Literal, overload

from numpy import int32
from numpy.typing import NDArray

from imas.ids_structure import IDSStructure

from ....ggd.unstruct_2d_extend_mesh import UnstructGrid2DExtended
from ....ggd.unstruct_2d_mesh import UnstructGrid2D
from .load_unstruct_2d import load_unstruct_grid_2d
from .load_unstruct_3d import load_unstruct_grid_2d_extended

__all__ = ["load_grid"]


@overload
def load_grid(
    grid_ggd: IDSStructure, with_subsets: Literal[False] = False, num_toroidal: int | None = None
) -> UnstructGrid2D | UnstructGrid2DExtended: ...


@overload
def load_grid(
    grid_ggd: IDSStructure, with_subsets: Literal[True], num_toroidal: int | None = None
) -> tuple[UnstructGrid2D, dict[str, NDArray[int32]], dict[str, int]]: ...


def load_grid(
    grid_ggd: IDSStructure, with_subsets: bool = False, num_toroidal: int | None = None
) -> (
    UnstructGrid2D
    | tuple[UnstructGrid2D, dict[str, NDArray[int32]], dict[str, int]]
    | UnstructGrid2DExtended
):
    """Load grid from the ``grid_ggd`` structure.

    The ``grid_ggd`` structure is expected to follow the IMAS GGD grid definition.
    Please see: https://imas-data-dictionary.readthedocs.io/en/latest/ggd_guide/doc.html#the-grid-ggd-aos

    .. warning::
        This function currently supports only unstructured 2D grids and unstructured 2D grids
        extended in 3D (mainly used in JOREK).
        Loading of structured grids and unstructured 3D grids will be implemented in the future.

    Parameters
    ----------
    grid_ggd
        The ``grid_ggd`` structure.
    with_subsets
        Read grid subset data, by default is False.
    num_toroidal
        Number of toroidal points, by default None.

    Returns
    -------
    grid : `.UnstructGrid2D` | `.UnstructGrid2DExtended`
        Grid object that depends on the grid type (2D/2D-extended/3D).
    subsets : `dict[str, NDArray[int32]]`
        Dictionary with grid subsets for each subset name containing the indices of the cells from
        that subset. Note that 'Cells' subset is included only if cell indices are specified.
    subset_id : `dict[str, int]`
        Dictionary with grid subset indices.

    Raises
    ------
    RuntimeError
        If the grid type is unsupported or if no spaces are found in the ``grid_ggd`` structure.
    NotImplementedError
        If the grid type is not yet implemented.

    Examples
    --------
    .. code-block:: python

        from imas import DBEntry
        from cherab.imas.ids.common import get_ids_time_slice
        from cherab.imas.ids.common.ggd import load_grid

        with DBEntry("imas:hdf5?path=/work/imas/shared/imasdb/ITER/4/123356/1", "r") as entry:
            ids = get_ids_time_slice(entry, "edge_profiles", 0)

        grid, subsets, subset_id = load_grid(ids.grid_ggd[0], with_subsets=True)
    """
    spaces = get_standard_spaces(grid_ggd)

    if not len(spaces):
        raise RuntimeError("GGD grid contain no spaces.")

    if len(spaces) == 1:  # simple unstructured grids
        if len(spaces[0].objects_per_dimension) == 3:  # 2D case
            return load_unstruct_grid_2d(grid_ggd, 0, with_subsets=with_subsets)
        if len(spaces[0].objects_per_dimension) == 4:  # 3D case
            raise NotImplementedError(
                "Loading unstructured 3D grids will be implemented in the future."
            )

        raise RuntimeError("Unsupported grid type.")

    if len(spaces) == 2:  # 2D structured grid or 2D unstructured grid extended in 3D
        if len(spaces[0].objects_per_dimension) == 3 and len(spaces[1].objects_per_dimension) < 3:
            return load_unstruct_grid_2d_extended(
                grid_ggd, with_subsets=with_subsets, num_toroidal=num_toroidal
            )
        if len(spaces[0].objects_per_dimension) < 3 and len(spaces[1].objects_per_dimension) < 3:
            raise NotImplementedError(
                "Loading structured 2D grids will be implemented in the future."
            )

        raise RuntimeError("Unsupported grid type.")

    if len(spaces) == 3:  # 3D structured grid
        if (
            len(spaces[0].objects_per_dimension) < 3
            and len(spaces[1].objects_per_dimension) < 3
            and len(spaces[2].objects_per_dimension) < 3
        ):
            raise NotImplementedError(
                "Loading structured 3D grids will be implemented in the future."
            )

        raise RuntimeError("Unsupported grid type.")

    raise RuntimeError("Unsupported grid type.")


def get_standard_spaces(grid_ggd: IDSStructure) -> list[IDSStructure]:
    """Get a list of standard non-empty spaces from the grid_ggd structure.

    Parameters
    ----------
    grid_ggd
        The grid_ggd structure.

    Returns
    -------
    list[IDSStructure]
        List of standard spaces.

    Raises
    ------
    ValueError
        If no spaces are defined in the grid_ggd structure.
    """
    if not len(grid_ggd.space):
        error_massage = "Unable to read the grid. Grid space is not defined."
        if len(grid_ggd.path):
            error_massage += f" The grid is defined in {grid_ggd.path}."
        raise ValueError(error_massage)

    # Get list of standard spaces:
    spaces = []
    for space in grid_ggd.space:
        # JINTRAC confuses geometry_type with identifier, uncomment this check when the bug is fixed
        # if space.geometry_type.index == 0 or space.geometry_type.index == EMPTY_INT:
        if len(space.objects_per_dimension) and len(space.objects_per_dimension[0].object) > 2:
            spaces.append(space)

    return spaces



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for loading bolometer cameras from IMAS bolometer IDS."""

from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass, field

import numpy as np
from numpy.typing import NDArray
from raysect.core.math import Point3D, Vector3D, from_cylindrical

from imas.ids_structure import IDSStructure
from imas.ids_toplevel import IDSToplevel

from .utility import CameraType, GeometryType


@dataclass
class Geometry:
    """Represent the geometric specification of a bolometer sensor head or slit aperture.

    The Geometry describes both simple rectangular / circular sensor faces and
    arbitrary polygonal / polyline definitions via explicit coordinates.
    It encapsulates a local orthonormal (or user supplied) basis, extents, and derived surface
    properties.
    """

    centre: Point3D = field(default_factory=Point3D)
    """Geometric centre (reference point) of the sensor/slit in global 3D coordinates."""
    type: GeometryType = GeometryType.RECTANGLE
    """Enumerated shape type.

    Common values include `RECTANGLE`, `CIRCLE`, `OUTLINE`, etc.
    Defaults to `.GeometryType.RECTANGLE`.
    """
    basis_x: Vector3D = field(default_factory=lambda: Vector3D(1, 0, 0))
    """Local x-axis direction vector lying in the sensor/slit plane."""
    basis_y: Vector3D = field(default_factory=lambda: Vector3D(0, 1, 0))
    """Local y-axis direction vector lying in the sensor/slit plane."""
    basis_z: Vector3D = field(default_factory=lambda: Vector3D(0, 0, 1))
    """Local outward-facing normal vector perpendicular to the sensor/slit plane.

    This vector must be directed toward the radiation sources.
    When None, it can be derived as the cross product of `basis_x` and `basis_y`.
    """
    dx: float = 0.0
    """Width along `basis_x` for rectangular geometry.

    None when not applicable (e.g., circular or polygonal types).
    """
    dy: float = 0.0
    """Width along `basis_y` for rectangular geometry.

    None when not applicable (e.g., circular or polygonal types).
    """
    surface: float = 0.0
    """Precomputed surface area of the aperture face.

    If None, may be derived from `dx`/`dy` (rectangles), `radius` (circles), or `coords` (polygons)
    during validation or runtime.
    """
    radius: float = 0.0
    """Radius for circular geometry types.

    None if the geometry is not circular.
    """
    coords: NDArray[np.float64] | None = None
    """Coordinate array defining the outline in the `basis_x` and `basis_y` plane.

    This array is used to represent a complex geometric outline.
    The array shape is ``(2, N)`` where ``N`` is the number of points.
    """


@dataclass
class BoloChannel:
    """Represent a bolometer camera channel.

    Each channel contains one foil and associated several slits.
    """

    foil: Geometry
    """Geometry of the foil used in the bolometer channel."""
    slits: list[Geometry]
    """List of geometries representing the slits associated with the bolometer channel."""
    num_subcollimator: int
    """Number of sub-collimators in the bolometer channel."""
    thickness_subcollimator: float
    """Thickness of each sub-collimator in the bolometer channel [m]."""


@dataclass
class BoloCamera:
    """Represent a bolometer camera and its associated data.

    This class encapsulates all the properties and data channels associated with
    a bolometer camera used for plasma diagnostics.
    """

    name: str
    """Unique identifier or label for the bolometer camera."""
    description: str
    """Detailed description of the bolometer camera, including its purpose, location, or other relevant information."""
    type: CameraType
    """Type of Bolometer camera: Pinhole/Collimator"""
    channels: list[BoloChannel]
    """List of individual bolometer channels belonging to this camera."""

    def __repre__(self) -> str:
        return f"<BoloCamera name={self.name} type={self.type.name} channels={len(self.channels)}>"

    def __len__(self) -> int:
        """Return the number of channels in the bolometer camera.

        Returns
        -------
        int
            Number of channels.
        """
        return len(self.channels)

    def __iter__(self) -> Iterator[BoloChannel]:
        """Return an iterator over the bolometer camera channels.

        Returns
        -------
        `Iterator[BoloChannel]`
            An iterator over the bolometer camera channels.
        """
        return iter(self.channels)

    def __getitem__(self, index: int) -> BoloChannel:
        """Return the bolometer channel at the specified index.

        Parameters
        ----------
        index
            Index of the channel to retrieve.

        Returns
        -------
        `.BoloChannel`
            The bolometer channel at the specified index.
        """
        return self.channels[index]


def load_cameras(ids: IDSToplevel) -> list[BoloCamera]:
    """Load bolometer cameras from the bolometer IDS.

    This function retrieves the camera information from the bolometer IDS and organizes it into a
    structured format.

    Parameters
    ----------
    ids
        The bolometer IDS.

    Returns
    -------
    `.BoloCamera`
        A list of bolometer camera data structures.

    Raises
    ------
    ValueError
        If the provided IDS is not a bolometer IDS.
    RuntimeError
        If no cameras are found in the IDS.
    """
    if not str(ids.metadata.name) == "bolometer":
        raise ValueError(f"Invalid bolometer IDS ({ids.metadata.name}).")

    cameras = getattr(ids, "camera", [])
    if not len(cameras):
        raise RuntimeError("No camera found in IDS.")

    bolo_data: list[BoloCamera] = []

    for camera in cameras:
        name = str(camera.name)
        description = str(camera.description)
        camera_type = CameraType.from_value(camera.type.index.value)

        channels: list[BoloChannel] = []
        for channel in camera.channel:
            # Sub-collimator
            n_subcollimators = int(channel.subcollimators_n.value)
            thickness_subcollimator = float(channel.subcollimators_separation.value)

            channels.append(
                BoloChannel(
                    foil=load_geometry(channel.detector),
                    slits=[load_geometry(aperture) for aperture in channel.aperture],
                    num_subcollimator=n_subcollimators if n_subcollimators > 1 else 1,
                    thickness_subcollimator=thickness_subcollimator
                    if thickness_subcollimator > 0
                    else 0.0,
                )
            )

        bolo_data.append(
            BoloCamera(
                name=name,
                description=description,
                type=camera_type,
                channels=channels,
            )
        )

    return bolo_data


def load_geometry(sensor: IDSStructure) -> Geometry:
    """Load the geometry of a sensor or aperture from the bolometer IDS.

    Parameters
    ----------
    sensor
        Detector or aperture structure object.

    Returns
    -------
    `.Geometry`
        Object with the following attributes:

        - ``'centre'``: Point3D with the coordinates of the sensor centre,
        - ``'type'``: Geometry type of the sensor,
        - ``'basis_x'``: Vector3D with the x-basis vector of the sensor,
        - ``'basis_y'``: Vector3D with the y-basis vector of the sensor,
        - ``'basis_z'``: Vector3D with the z-basis vector of the sensor,
        - ``'dx'``: Width of the sensor in the x-direction [m],
        - ``'dy'``: Width of the sensor in the y-direction [m],
        - ``'surface'``: Surface area of the sensor [m²],
        - ``'radius'``: Radius of the sensor [m] if the geometry type is circular,
        - ``'coords'``: Outline coordinates for basis_x and basis_y if the geometry type is outline.

        Some keys may not be present depending on the geometry type.

    Raises
    ------
    ValueError
        If the geometry type is invalid.
    """
    geometry = Geometry()

    geometry.centre = from_cylindrical(
        sensor.centre.r, sensor.centre.z, np.rad2deg(sensor.centre.phi)
    )

    geometry_type = GeometryType.from_value(int(sensor.geometry_type.value))
    geometry.type = geometry_type

    match geometry_type:
        case GeometryType.RECTANGLE:
            # Unit vectors
            geometry.basis_x = Vector3D(
                sensor.x2_unit_vector.x, sensor.x2_unit_vector.y, sensor.x2_unit_vector.z
            )
            geometry.basis_y = Vector3D(
                sensor.x1_unit_vector.x, sensor.x1_unit_vector.y, sensor.x1_unit_vector.z
            )
            geometry.basis_z = Vector3D(
                sensor.x3_unit_vector.x, sensor.x3_unit_vector.y, sensor.x3_unit_vector.z
            )

            # Dimensions
            geometry.dx = float(sensor.x2_width.value)
            geometry.dy = float(sensor.x1_width.value)

        case GeometryType.CIRCULAR:
            # Radius
            radius = getattr(sensor, "radius", None)
            if radius is None or radius <= 0:
                raise ValueError(f"Invalid radius ({radius}).")

            geometry.radius = float(radius.value)

            # Unit vectors
            geometry.basis_z = Vector3D(
                sensor.x3_unit_vector.x, sensor.x3_unit_vector.y, sensor.x3_unit_vector.z
            )
            geometry.basis_x = geometry.basis_z.orthogonal().normalise()
            geometry.basis_y = geometry.basis_z.cross(geometry.basis_x).normalise()

        case GeometryType.OUTLINE:
            # Outline coordinates for basis_x and basis_y
            geometry.coords = np.vstack((sensor.outline.x2, sensor.outline.x1))

            # Unit vectors
            geometry.basis_x = Vector3D(
                sensor.x2_unit_vector.x, sensor.x2_unit_vector.y, sensor.x2_unit_vector.z
            )
            geometry.basis_y = Vector3D(
                sensor.x1_unit_vector.x, sensor.x1_unit_vector.y, sensor.x1_unit_vector.z
            )
            geometry.basis_z = Vector3D(
                sensor.x3_unit_vector.x, sensor.x3_unit_vector.y, sensor.x3_unit_vector.z
            )

    # Surface area
    surface = getattr(sensor, "surface", None)
    geometry.surface = float(surface.value) if surface is not None else 0.0

    return geometry



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for common functions used to get IDS species information."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum

import numpy as np
from numpy.typing import NDArray

from cherab.core.atomic.elements import Element, Isotope, lookup_isotope
from imas.ids_defs import EMPTY_FLOAT, EMPTY_INT
from imas.ids_struct_array import IDSStructArray
from imas.ids_structure import IDSStructure

__all__ = [
    "SpeciesType",
    "SpeciesData",
    "ProfileData",
    "SpeciesComposition",
    "VelocityData",
    "get_ion_state",
    "get_neutral_state",
    "get_ion",
    "get_neutral",
    "get_elements",
]


class SpeciesType(Enum):
    """Enumeration of species types in IMAS."""

    ION = "ion"
    """Single ion state, with a specific ionization state (z_min == z_max)"""
    ION_BUNDLE = "ion_bundle"
    """Bundle of ion states, with a range of ionization states (z_min != z_max)"""
    NEUTRAL = "neutral"
    """Single neutral state"""
    NEUTRAL_BUNDLE = "neutral_bundle"
    """Bundle of neutral states"""
    MOLECULE = "molecule"
    """Single molecule state"""
    MOLECULAR_BUNDLE = "molecular_bundle"
    """Bundle of molecular states"""


@dataclass
class SpeciesData:
    """Dataclass to represent the data of a species in IMAS."""

    z_min: int
    """Minimum ionization state of the species"""
    z_max: int
    """Maximum ionization state of the species"""
    element: Element | None = None
    """Element that makes up the species, if it is a single particle"""
    elements: tuple[Element, ...] = field(default_factory=tuple)
    """Elements that make up the species, if it is a molecule"""
    species_type: SpeciesType | None = None
    """Type of species"""
    electron_configuration: str | None = None
    """Electron configuration of the species"""
    vibrational_mode: str | None = None
    """Vibrational mode of the species, if it is a molecule"""
    vibrational_level: float | None = None
    """Vibrational level of the species, if it is a molecule"""

    def __str__(self) -> str:
        """Return a string representation of the species data.

        Returns
        -------
        str
            String representation of the species data.
        """
        if self.species_type in {SpeciesType.ION, SpeciesType.NEUTRAL}:
            if self.element is not None:
                return f"{self.element.symbol} {self.species_type.value} (z=+{self.z_min})"
            else:
                return f"{self.species_type.value} (z=+{self.z_min})"
        elif self.species_type in {SpeciesType.ION_BUNDLE, SpeciesType.NEUTRAL_BUNDLE}:
            if self.element is not None:
                return (
                    f"{self.element.symbol} {self.species_type.value} (z={self.z_min}-{self.z_max})"
                )
            else:
                return f"{self.species_type.value} (z={self.z_min}-{self.z_max})"
        elif self.species_type == SpeciesType.MOLECULE:
            return f"{'-'.join(el.symbol for el in self.elements)} {self.species_type.value}"
        elif self.species_type == SpeciesType.MOLECULAR_BUNDLE:
            return f"{'-'.join(el.symbol for el in self.elements)} {self.species_type.value} (z={self.z_min}-{self.z_max})"
        else:
            return "Unknown species type"


@dataclass
class VelocityData:
    """Dataclass for storing the bulk velocity data of a species."""

    radial: NDArray[np.float64] | None = None
    """Radial velocity [m/s]."""
    parallel: NDArray[np.float64] | None = None
    """Parallel velocity [m/s]."""
    poloidal: NDArray[np.float64] | None = None
    """Poloidal velocity [m/s]."""
    r: NDArray[np.float64] | None = None
    """Radial velocity along the major radius axis [m/s]."""
    phi: NDArray[np.float64] | None = None
    """Toroidal velocity [m/s]."""
    z: NDArray[np.float64] | None = None
    """Vertical velocity along the height axis [m/s]."""


@dataclass
class ProfileData:
    """Dataclass for storing the profile data of a species."""

    species: SpeciesData
    """Data of the species."""
    density: NDArray[np.float64] | None = None
    """Density (thermal+non-thermal) [m^-3]."""
    density_thermal: NDArray[np.float64] | None = None
    """Density (thermal) [m^-3]."""
    density_fast: NDArray[np.float64] | None = None
    """Density of fast (non-thermal) particles [m^-3]."""
    temperature: NDArray[np.float64] | None = None
    """Temperature [eV]."""
    velocity: VelocityData | None = None
    """Bulk velocity data of the species."""


@dataclass
class SpeciesComposition:
    """Dataclass for storing the composition of the plasma species."""

    electron: ProfileData
    """Electron profiles."""
    ion: list[ProfileData] = field(default_factory=list)
    """Ion profiles."""
    ion_bundle: list[ProfileData] = field(default_factory=list)
    """Ion bundle profiles."""
    neutral: list[ProfileData] = field(default_factory=list)
    """Neutral particle profiles."""
    neutral_bundle: list[ProfileData] = field(default_factory=list)
    """Neutral bundle profiles."""
    molecule: list[ProfileData] = field(default_factory=list)
    """Molecule profiles."""
    molecular_bundle: list[ProfileData] = field(default_factory=list)
    """Molecular bundle profiles."""


def get_ion_state(
    state: IDSStructure,
    state_index: int,
    elements: tuple[Element, ...],
    grid_subset_index: int | None = None,
) -> SpeciesData:
    """Get a unique identifier for an ion state.

    Parameters
    ----------
    state
        IDSStructure representing `.../ion[i]/state`
    state_index
        Index of the state in the list of states.
    elements
        Tuple of elements that make up the ion state.
    grid_subset_index
        The grid subset index to use for 1D profiles, by default None.

    Returns
    -------
    `.SpeciesData`
        Instance of the `SpeciesData` dataclass representing the ion state.
    """
    if state.z_min == EMPTY_FLOAT or state.z_max == EMPTY_FLOAT:
        if grid_subset_index is None:  # 1D profiles
            z_average = state.z_average_1d
        else:
            for s in state.z_average:
                if s.grid_subset_index == grid_subset_index:
                    z_average = s.values
                    break
            else:
                z_average = []
        if len(z_average):  # probably, a bundle
            z_min = (
                int(state.z_min) if state.z_min != EMPTY_FLOAT else int(np.floor(min(z_average)))
            )
            z_max = int(state.z_max) if state.z_max != EMPTY_FLOAT else int(np.ceil(max(z_average)))
        else:  # probably, a single ion
            z_min = int(state.z_min) if state.z_min != EMPTY_FLOAT else state_index + 1
            z_max = int(state.z_max) if state.z_max != EMPTY_FLOAT else z_min
    else:
        z_min = int(state.z_min)
        z_max = int(state.z_max)

    # Initialize the state species dataclass
    species_data = SpeciesData(
        z_min=z_min,
        z_max=z_max,
        electron_configuration=str(getattr(state, "electron_configuration", "")).strip()
        if len(getattr(state, "electron_configuration", "")) > 0
        else None,
    )

    if len(elements) > 1:  # molecular ions and bundles
        species_data.elements = elements
        if z_min == z_max == 0:
            species_data.species_type = SpeciesType.NEUTRAL_BUNDLE
        elif z_min == z_max:
            species_data.species_type = SpeciesType.MOLECULE
            species_data.vibrational_mode = (
                str(state.vibrational_mode) if len(state.vibrational_mode) else None
            )
            species_data.vibrational_level = (
                state.vibrational_level if state.vibrational_level != EMPTY_FLOAT else None
            )
        else:
            species_data.species_type = SpeciesType.MOLECULAR_BUNDLE
    else:  # ions and bundles
        species_data.element = elements[0]
        if z_min == z_max == 0:
            species_data.species_type = SpeciesType.NEUTRAL
        elif z_min == z_max:
            species_data.species_type = SpeciesType.ION
        else:
            species_data.species_type = SpeciesType.ION_BUNDLE

    return species_data


def get_neutral_state(state: IDSStructure, elements: tuple[Element, ...]) -> SpeciesData:
    """Get a unique identifier for a neutral state.

    Parameters
    ----------
    state
        The neutral_state structure from IMAS.
    elements
        Tuple of elements that make up the neutral state.

    Returns
    -------
    `.SpeciesData`
        Instance of the `SpeciesData` dataclass representing the neutral state.
    """
    # Initialize the state species dataclass
    species_data = SpeciesData(
        z_min=0,
        z_max=0,
        electron_configuration=str(getattr(state, "electron_configuration", "")).strip()
        if len(getattr(state, "electron_configuration", "")) > 0
        else None,
    )
    if len(elements) > 1:  # molecules and bundles
        species_data.elements = elements
        if (
            getattr(state, "vibrational_mode", None)
            and getattr(state, "vibrational_level", EMPTY_FLOAT) != EMPTY_FLOAT
        ):
            species_data.species_type = SpeciesType.MOLECULE
            species_data.vibrational_mode = (
                str(getattr(state, "vibrational_mode", "")).strip()
                if len(getattr(state, "vibrational_mode", ""))
                else None
            )
            species_data.vibrational_level = (
                getattr(state, "vibrational_level", EMPTY_FLOAT)
                if getattr(state, "vibrational_level", EMPTY_FLOAT) != EMPTY_FLOAT
                else None
            )
        else:
            species_data.species_type = SpeciesType.NEUTRAL_BUNDLE
    else:  # neutrals
        species_data.element = elements[0]
        species_data.species_type = SpeciesType.NEUTRAL

    return species_data


def get_ion(ion: IDSStructure, elements: tuple[Element, ...]) -> SpeciesData:
    """Get a unique identifier for an ion or molecule.

    Parameters
    ----------
    ion
        The ion structure from IMAS.
    elements
        Tuple of elements that make up the ion.

    Returns
    -------
    `.SpeciesData`
        Instance of the `SpeciesData` dataclass representing the ion or molecule.
    """
    z_ion = int(ion.z_ion) if ion.z_ion != EMPTY_FLOAT else elements[0].atomic_number
    species_data = SpeciesData(
        z_min=z_ion,
        z_max=z_ion,
        element=elements[0] if len(elements) == 1 else None,
        elements=elements if len(elements) > 1 else tuple(),
        species_type=SpeciesType.MOLECULE if len(elements) > 1 else SpeciesType.ION,
    )

    return species_data


def get_neutral(neutral: IDSStructure, elements: tuple[Element, ...]) -> SpeciesData:
    """Get a unique identifier for a neutral or molecule.

    Parameters
    ----------
    neutral
        The neutral structure from IMAS.
    elements
        Tuple of elements that make up the neutral state.

    Returns
    -------
    `.SpeciesData`
        Instance of the `SpeciesData` dataclass representing the neutral or molecule.
    """
    species_data = SpeciesData(
        z_min=0,
        z_max=0,
        element=elements[0] if len(elements) == 1 else None,
        elements=elements if len(elements) > 1 else tuple(),
        species_type=SpeciesType.MOLECULE if len(elements) > 1 else SpeciesType.NEUTRAL,
    )

    return species_data


def get_elements(elements_aos: IDSStructArray) -> tuple[Element | Isotope, ...]:
    """Get a tuple of elements from an IDS array of structures.

    Parameters
    ----------
    elements_aos
        Element IDS array of structures

    Returns
    -------
    tuple[`~cherab.core.atomic.elements.Element` | `~cherab.core.atomic.elements.Isotope`, ...]
        Tuple of elements that make up the species, with isotopes preferred over elements when possible.
    """
    elements = []
    for element in elements_aos:
        mass_number = int(round(element.a))
        zn = int(round(element.z_n))
        isotope = lookup_isotope(zn, number=mass_number)
        if int(round(isotope.element.atomic_weight)) == mass_number:
            # Prefer element over isotope
            isotope = isotope.element

        if getattr(element, "atoms_n", EMPTY_INT) == EMPTY_INT:
            atoms_n = 1
        else:
            atoms_n = int(round(getattr(element, "atoms_n", EMPTY_INT)))

        for _ in range(atoms_n):
            elements.append(isotope)

    return tuple(elements)



"""Module including patch functionalities for datasets."""

from cherab.imas.ids.common import get_ids_time_slice
from imas import DBEntry


def fix_jintrac(path_in: str, path_out: str) -> None:
    """Fix the JINTRAC IDS.

    This function modifies the JINTRAC IDS by updating the `z_min` and `z_max` values for the ion
    states of Neon and Tungsten in the `core_profiles` IDS.

    The modified IDS is then saved to a new file with a "_mod" suffix.

    Parameters
    ----------
    path_in
        The file path to the original JINTRAC IDS.
    path_out
        The file path to save the modified IDS.
    """
    print("=== Apply patch to fix ===")
    # %%
    # Define the bundles for Neon and Tungsten
    # ----------------------------------------
    BUNDLES_W = [
        (1, 1),
        (2, 6),
        (7, 12),
        (13, 22),
        (23, 73),
        (74, 74),
    ]
    BUNDLES_NE = [
        (1, 1),
        (2, 3),
        (4, 6),
        (7, 9),
        (10, 10),
    ]

    # %%
    # Load the IDS from the JINTRAC simulation
    # ----------------------------------------
    with DBEntry(path_in, "r+") as entry:
        ids_core = get_ids_time_slice(entry, "core_profiles", 0.0)
        ids_others = [
            entry.get("edge_profiles", autoconvert=False),
            entry.get("core_sources", autoconvert=False),
            entry.get("edge_sources", autoconvert=False),
            entry.get("edge_transport", autoconvert=False),
            entry.get("equilibrium", autoconvert=False),
            entry.get("ntms", autoconvert=False),
            entry.get("summary", autoconvert=False),
        ]

    # %%
    # Update the z_min and z_max values for the ion states based on the defined bundles
    # ---------------------------------------------------------------------------------
    for i_ion in [3, 4]:
        ion = ids_core.profiles_1d[0].ion[i_ion]
        bundles = BUNDLES_NE if i_ion == 3 else BUNDLES_W
        for i, state in enumerate(ion.state):
            state.z_min = bundles[i][0]
            state.z_max = bundles[i][1]

    # %%
    # Save the modified IDS to a new file
    # -----------------------------------
    with DBEntry(path_out, "w", dd_version=ids_core._version) as entry:
        entry.put(ids_core)
        for ids in ids_others:
            entry.put(ids)



# cython: language_level=3

# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module defining unit vector functions."""

from raysect.core.math cimport Vector3D
from raysect.core.math.function.vector3d cimport autowrap_function1d as autowrap_vectorfunction1d
from raysect.core.math.function.vector3d cimport autowrap_function2d as autowrap_vectorfunction2d
from raysect.core.math.function.vector3d cimport autowrap_function3d as autowrap_vectorfunction3d

__all__ = ["UnitVector1D", "UnitVector2D", "UnitVector3D"]


cdef class UnitVector1D(VectorFunction1D):
    """Evaluate a unit vector for the given VectorFunction1D instance."""

    def __init__(self, object vector):
        self._vector = autowrap_vectorfunction1d(vector)

    cdef Vector3D evaluate(self, double x):

        return self._vector.evaluate(x).normalise()


cdef class UnitVector2D(VectorFunction2D):
    """Evaluate a unit vector for the given VectorFunction2D instance."""

    def __init__(self, object vector):
        self._vector = autowrap_vectorfunction2d(vector)

    cdef Vector3D evaluate(self, double x, double y):

        return self._vector.evaluate(x, y).normalise()


cdef class UnitVector3D(VectorFunction3D):
    """Evaluate a unit vector for the given VectorFunction3D instance."""

    def __init__(self, object vector):
        self._vector = autowrap_vectorfunction3d(vector)

    cdef Vector3D evaluate(self, double x, double y, double z):

        return self._vector.evaluate(x, y, z).normalise()



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module defining unstructured 2D-extended mesh class and related methods."""

from __future__ import annotations

import sys
from typing import Literal

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override  # pyright: ignore[reportUnreachable]

import matplotlib.axes
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.collections import PolyCollection
from matplotlib.tri import Triangulation
from numpy.typing import ArrayLike, NDArray
from raysect.core.math.vector import Vector3D

from ..math import UnstructGridFunction3D, UnstructGridVectorFunction3D
from ..math.tetrahedralize import calculate_tetra_volume, cell_to_5tetra
from .base_mesh import GGDGrid

__all__ = ["UnstructGrid2DExtended"]

ZERO_VECTOR = Vector3D(0, 0, 0)


class UnstructGrid2DExtended(GGDGrid):
    """Unstructured 2D grid with toroidal extension.

    The grid cells are voxels (cubes). Vertices may be shared with neighbouring cells.

    To use Raysect's KDtree accelerator, each cell is tetrahedralized.

    Parameters
    ----------
    vertices
        Array-like of shape ``(N, 3)`` containing coordinates of the cell vertices in the (X, Y, Z)
        space.
    cells
        Array-like of shape ``(M, 8)`` containing the vertex indices in clockwise or
        counterclockwise order for each cubic cell in the list
        (the starting vertex must not be included twice).
    num_faces
        Number of faces at the poloidal plane.
    num_poloidal
        Number of poloidal points.
    num_toroidal
        Number of toroidal points.
    name
        Name of the grid, by default ``"Cells"``.
    coordinate_system
        Coordinate system of the grid, by default ``"cylindrical"``.
    """

    def __init__(
        self,
        vertices: ArrayLike,
        cells: ArrayLike,
        num_faces: int,
        num_poloidal: int,
        num_toroidal: int,
        name: str = "Cells",
        coordinate_system: Literal["cylindrical", "cartesian"] = "cylindrical",
    ):
        vertices = np.array(vertices, dtype=np.float64)
        vertices.setflags(write=False)
        cells = np.array(cells, dtype=np.int32)
        cells.setflags(write=False)

        if vertices.ndim != 2:
            raise ValueError(
                "Attribute 'vertices' must be a 2D array-like. "
                + f"The number of dimensions in 'vertices' is {vertices.ndim}."
            )

        if vertices.shape[1] != 3:
            raise ValueError(
                "Attribute 'vertices' must be a (N, 3) array-like. "
                + f"The shape of 'vertices' is {vertices.shape}."
            )

        if cells.ndim != 2:
            raise ValueError(
                "Attribute 'cells' must be a 2D array-like. "
                + f"The number of dimensions in 'cells' is {cells.ndim}."
            )

        if cells.shape[1] != 8:
            raise ValueError(
                "Attribute 'cells' must be a (M, 8) array-like. "
                + f"The shape of 'cells' is {cells.shape}."
            )

        self._vertices: NDArray[np.float64] = vertices
        self._cells: NDArray[np.int32] = cells
        self._num_faces: int = num_faces
        self._num_poloidal: int = num_poloidal
        self._num_toroidal: int = num_toroidal

        super().__init__(name, 3, coordinate_system)

    @override
    def _initial_setup(self):
        self._interpolator = None

        self._num_cell: int = self._cells.shape[0]

        # Extract grid points at the first poloidal plane
        r = self._vertices[: self._num_poloidal, 0]
        z = self._vertices[: self._num_poloidal, 2]

        # Work out the extent of the mesh (both Cartesian and cylindrical).
        self._mesh_extent = {
            "xmin": self._vertices[:, 0].min(),
            "xmax": self._vertices[:, 0].max(),
            "ymin": self._vertices[:, 1].min(),
            "ymax": self._vertices[:, 1].max(),
            "rmin": r.min(),
            "rmax": r.max(),
            "zmin": z.min(),
            "zmax": z.max(),
        }

        # Tetrahedralize cells
        self._tetrahedra = cell_to_5tetra(self._cells)
        self._tetra_to_cell_map = np.repeat(np.arange(len(self._cells), dtype=np.int32), 5)
        self._cell_to_tetra_map = np.column_stack(
            (np.arange(len(self._cells)) * 5, np.full(len(self._cells), 5))
        )

        self._tetrahedra.setflags(write=False)
        self._cell_to_tetra_map.setflags(write=False)
        self._tetra_to_cell_map.setflags(write=False)

        # Calculate cell volume and centroid
        self._cell_volume = np.empty(len(self._cells), dtype=np.float64)
        self._cell_centre = np.empty((len(self._cells), 3), dtype=np.float64)

        for i, cell in enumerate(self._cells):
            vertices = self._vertices[cell]
            self._cell_centre[i] = vertices.mean(axis=0)
            itet, ntet = self._cell_to_tetra_map[i]
            self._cell_volume[i] = calculate_tetra_volume(
                self._vertices, self._tetrahedra[itet : itet + ntet]
            )

        self._cell_volume.setflags(write=False)
        self._cell_centre.setflags(write=False)

    @property
    def num_poloidal(self) -> int:
        """Number of poloidal grid points."""
        return self._num_poloidal

    @property
    def num_toroidal(self) -> int:
        """Number of toroidal grid points."""
        return self._num_toroidal

    @property
    def num_faces(self) -> int:
        """Number of faces at the poloidal plane."""
        return self._num_faces

    @property
    def vertices(self) -> NDArray[np.float64]:
        """Mesh vertex coordinates as ``(N, 3)`` array in the (X, Y, Z) space."""
        return self._vertices

    @property
    def cells(self) -> NDArray[np.int32]:
        """Mesh cells as ``(M, 8)`` array."""
        return self._cells

    @property
    def tetrahedra(self) -> NDArray[np.int32]:
        """Mesh tetrahedra as ``(5M, 4)`` array."""
        return self._tetrahedra

    @property
    def tetra_to_cell_map(self) -> NDArray[np.int32]:
        """Array of shape ``(5M,)`` mapping every tetrahedral index to a grid cell ID."""
        return self._tetra_to_cell_map

    @property
    def cell_to_tetra_map(self) -> NDArray[np.int32]:
        """Array of shape ``(M, 2)`` mapping every grid cell index to tetrahedral IDs.

        The first column is the index of the first tetrahedron forming the cell.
        The second column is the number of tetrahedra forming the cell.

        >>> itet, ntet = mesh.cell_to_tetra_map[icell]
        >>> tet_cell = mesh.tetrahedra[itet : itet + ntet]
        """
        return self._cell_to_tetra_map

    def subset_faces(self, indices: ArrayLike, name: str | None = None) -> UnstructGrid2DExtended:
        """Create a subset UnstructGrid2DExtended from this instance.

        The subset is defined by the indices of the faces at the poloidal plane.
        Thus, the subset mesh is periodic in the toroidal direction.

        Parameters
        ----------
        indices
            Indices of the faces of the original grid in the subset.
        name
            Name of the grid subset. Default is ``instance.name + " subset"``.

        Returns
        -------
        `.UnstructGrid2DExtended`
            Subset instance.

        Raises
        ------
        ValueError
            If any of the indices of the faces is out of range.

        See Also
        --------
        subset : For a method that creates a subset from the original grid cells.
        """
        grid = UnstructGrid2DExtended.__new__(UnstructGrid2DExtended)

        grid._name = name or self._name + " subset"
        grid._coordinate_system = self._coordinate_system
        grid._dimension = self._dimension
        grid._interpolator = None
        if np.amax(indices) >= self._num_faces:
            raise ValueError(
                "The indices of the faces must be less than the number of faces in the grid."
            )
        indices = np.unique(indices)
        grid._num_faces = len(indices)
        grid._num_toroidal = self._num_toroidal

        index_flags: NDArray[np.bool_] = np.zeros(self._num_faces, dtype=bool)
        index_flags[indices] = True
        index_flags = np.tile(index_flags, self._num_toroidal)
        index_flags.setflags(write=False)

        # TODO: Require to implement index_flags as a property neatly?
        grid.index_flags = index_flags

        cells_original = self.cells[
            index_flags
        ]  # all cells in this subset but with original vertex indices
        vert_index, inv_index = np.unique(
            cells_original, return_inverse=True
        )  # all unique vertex indices in this subset
        grid._vertices = np.array(self.vertices[vert_index])  # vertices in this subset
        grid._vertices.setflags(write=False)

        # renumerating vertex indices
        cells = []  # and split
        ist = 0
        for cell in cells_original:
            cells.append(inv_index[ist : ist + len(cell)])
            ist += len(cell)
        grid._cells = np.vstack(cells, dtype=np.int32)
        grid._cells.setflags(write=False)
        grid._num_cell = len(grid._cells)
        grid._num_poloidal = grid._vertices.shape[0] // self._num_toroidal

        # cell volume and centres of this subset
        grid._cell_volume = np.array(self._cell_volume[index_flags])
        grid._cell_volume.setflags(write=False)
        grid._cell_centre = np.array(self._cell_centre[index_flags])
        grid._cell_centre.setflags(write=False)

        # Extract grid points at the first poloidal plane
        r = np.hypot(grid._vertices[:, 0], grid._vertices[:, 1])
        z = grid._vertices[:, 2]

        # mesh extent of this subset
        grid._mesh_extent = {
            "xmin": grid._vertices[:, 0].min(),
            "xmax": grid._vertices[:, 0].max(),
            "ymin": grid._vertices[:, 1].min(),
            "ymax": grid._vertices[:, 1].max(),
            "rmin": r.min(),
            "rmax": r.max(),
            "zmin": z.min(),
            "zmax": z.max(),
        }

        # Tetrahedralize cells and maps of this subset
        grid._tetrahedra = cell_to_5tetra(grid._cells)
        grid._tetra_to_cell_map = np.repeat(np.arange(len(grid._cells), dtype=np.int32), 5)
        grid._cell_to_tetra_map = np.column_stack(
            (np.arange(len(grid._cells)) * 5, np.full(len(grid._cells), 5))
        )

        grid._tetrahedra.setflags(write=False)
        grid._cell_to_tetra_map.setflags(write=False)
        grid._tetra_to_cell_map.setflags(write=False)

        return grid

    @override
    def subset(self, indices: ArrayLike, name: str | None = None) -> UnstructGrid2DExtended:
        """Create a subset UnstructGrid2DExtended from this instance.

        .. warning::
            The subset loses the range of cylindrical coordinates
            because the extracted vertex data is not necessarily periodic in the toroidal direction.

        Parameters
        ----------
        indices
            Indices of the cells of the original grid in the subset.
        name
            Name of the grid subset. Default is ``instance.name + " subset"``.

        Returns
        -------
        `.UnstructGrid2DExtended`
            Subset instance.
        """
        grid = UnstructGrid2DExtended.__new__(UnstructGrid2DExtended)

        grid._name = name or self._name + " subset"
        grid._coordinate_system = self._coordinate_system
        grid._dimension = self._dimension
        grid._interpolator = None
        grid._num_faces = None
        grid._num_poloidal = None
        grid._num_toroidal = None

        cells_original = self.cells[
            indices
        ]  # all cells in this subset but with original vertex indices
        vert_index, inv_index = np.unique(
            cells_original, return_inverse=True
        )  # all unique vertex indices in this subset
        grid._vertices = np.array(self.vertices[vert_index])  # vertices in this subset
        grid._vertices.setflags(write=False)

        # renumerating vertex indices
        cells = []  # and split
        ist = 0
        for cell in cells_original:
            cells.append(inv_index[ist : ist + len(cell)])
            ist += len(cell)
        grid._cells = np.vstack(cells, dtype=np.int32)
        grid._cells.setflags(write=False)
        grid._num_cell = len(grid._cells)

        # cell volume and centres of this subset
        grid._cell_volume = np.array(self._cell_volume[indices])
        grid._cell_volume.setflags(write=False)
        grid._cell_centre = np.array(self._cell_centre[indices])
        grid._cell_centre.setflags(write=False)

        # mesh extent of this subset
        grid._mesh_extent = {
            "xmin": grid._vertices[:, 0].min(),
            "xmax": grid._vertices[:, 0].max(),
            "ymin": grid._vertices[:, 1].min(),
            "ymax": grid._vertices[:, 1].max(),
        }

        # Tetrahedralize cells and maps of this subset
        grid._tetrahedra = cell_to_5tetra(grid._cells)
        grid._tetra_to_cell_map = np.repeat(np.arange(len(grid._cells), dtype=np.int32), 5)
        grid._cell_to_tetra_map = np.column_stack(
            (np.arange(len(grid._cells)) * 5, np.full(len(grid._cells), 5))
        )

        grid._tetrahedra.setflags(write=False)
        grid._cell_to_tetra_map.setflags(write=False)
        grid._tetra_to_cell_map.setflags(write=False)

        return grid

    @override
    def interpolator(
        self, grid_data: NDArray[np.float64], fill_value: float = 0
    ) -> UnstructGridFunction3D:
        """Return an UnstructGridFunction3D interpolator instance for the data defined on this grid.

        On the second and subsequent calls, the interpolator is created as an instance
        of the previously created interpolator sharing the same KDtree structure.

        Parameters
        ----------
        grid_data
            Array containing data in the grid cells.
        fill_value
            Value returned outside the grid, by default 0.

        Returns
        -------
        `.UnstructGridFunction3D`
            Interpolator instance.
        """
        if self._interpolator is None:
            self._interpolator = UnstructGridFunction3D(
                self._vertices, self._tetrahedra, self._tetra_to_cell_map, grid_data, fill_value
            )
            return self._interpolator

        return UnstructGridFunction3D.instance(self._interpolator, grid_data, fill_value)

    @override
    def vector_interpolator(
        self, grid_vectors: ArrayLike, fill_vector: Vector3D = ZERO_VECTOR
    ) -> UnstructGridVectorFunction3D:
        """Return an `UnstructGridVectorFunction3D` interpolator instance for the vector data defined on this grid.

        On the second and subsequent calls, the interpolator is created as an instance
        of the previously created interpolator sharing the same KDtree structure.

        Parameters
        ----------
        grid_vectors
            ``(3, L)`` Array containing 3D vectors in the grid cells.
        fill_vector
            3D vector returned outside the grid.

        Returns
        -------
        `.UnstructGridVectorFunction3D`
            Interpolator instance.
        """
        if self._interpolator is None:
            self._interpolator = UnstructGridVectorFunction3D(
                self._vertices,
                self._tetrahedra,
                self._tetra_to_cell_map,
                grid_vectors,
                fill_vector,
            )
            return self._interpolator

        return UnstructGridVectorFunction3D.instance(self._interpolator, grid_vectors, fill_vector)

    @override
    def __getstate__(self):
        """Get the state of the UnstructGrid2DExtended instance for serialization.

        Returns
        -------
        Dictionary with the instance attributes.
        """
        state = {
            "name": self._name,
            "dimension": self._dimension,
            "coordinate_system": self._coordinate_system,
            "vertices": self._vertices,
            "cells": self._cells,
            "num_poloidal": self._num_poloidal,
            "num_toroidal": self._num_toroidal,
            "num_faces": self._num_faces,
            "index_flags": self.index_flags if hasattr(self, "index_flags") else None,
        }
        return state

    def __setstate__(self, state):
        """Restore the state of the UnstructGrid2DExtended instance."""
        self._name = state["name"]
        self._dimension = state["dimension"]
        self._coordinate_system = state["coordinate_system"]
        self._vertices = state["vertices"]
        self._vertices.setflags(write=False)
        self._cells = state["cells"]
        self._cells.setflags(write=False)
        self._num_poloidal = state["num_poloidal"]
        self._num_toroidal = state["num_toroidal"]
        self._num_faces = state["num_faces"]
        if "index_flags" in state:
            # If index_flags exists, it is a subset of the original grid
            self.index_flags = state["index_flags"]

        self._initial_setup()

    def plot_tetra_mesh(
        self, data: ArrayLike | None = None, ax: matplotlib.axes.Axes | None = None
    ) -> None:
        """Plot the tetrahedral mesh grid geometry.

        .. warning::
            Plotting of tetrahedral mesh is not implemented yet.

        Parameters
        ----------
        data
            Data array defined on the tetrahedral mesh.
        ax
            Matplotlib axes to plot the mesh. If None, a new figure is created.

        Raises
        ------
        NotImplementedError
            Plotting of tetrahedral mesh is not implemented yet.
        """
        raise NotImplementedError("Plotting of tetrahedral mesh is not implemented yet.")

    @override
    def plot_mesh(
        self,
        data: ArrayLike | None = None,
        ax: matplotlib.axes.Axes | None = None,
        **grid_styles: str | float,
    ) -> matplotlib.axes.Axes:
        """Plot the polygonal mesh grid geometry at the first poloidal plane to a matplotlib figure.

        Parameters
        ----------
        data
            Data array defined on the polygonal mesh at the poloidal plane.
        ax
            Matplotlib axes to plot the mesh. If None, a new figure is created.
        **grid_styles
            Styles for the grid lines and faces,
            by default ``{"facecolor": "none", "edgecolor": "b", "linewidth": 0.25}``.

        Returns
        -------
        `~matplotlib.axes.Axes`
            Matplotlib axes with the plotted mesh.
        """
        if ax is None:
            _, ax = plt.subplots(layout="constrained")

        # Set default grid line styles if not provided
        grid_styles.setdefault("facecolor", "none")
        grid_styles.setdefault("edgecolor", "b")
        grid_styles.setdefault("linewidth", 0.25)

        cells = self._cells[: self._num_faces, :4]
        verts = [self._vertices[cell][:, ::2] for cell in cells]

        if data is None:
            collection_mesh = PolyCollection(verts, **grid_styles)
        else:
            collection_mesh = PolyCollection(verts)
            collection_mesh.set_array(data)
        ax.add_collection(collection_mesh)
        ax.set_aspect(1)
        ax.set_xlim(self._mesh_extent["rmin"], self._mesh_extent["rmax"])
        ax.set_ylim(self._mesh_extent["zmin"], self._mesh_extent["zmax"])

        if self._coordinate_system == "cartesian":
            ax.set_xlabel("X [m]")
            ax.set_ylabel("Y [m]")
        elif self._coordinate_system == "cylindrical":
            ax.set_xlabel("R [m]")
            ax.set_ylabel("Z [m]")

        return ax

    def plot_tri_mesh(
        self, data: ArrayLike, ax: matplotlib.axes.Axes | None = None, cmap: str = "viridis"
    ) -> matplotlib.axes.Axes:
        """Plot the data defined on the triangular mesh at the poloidal plane to a matplotlib figure.

        Parameters
        ----------
        data
            Data array defined on the triangular mesh at the poloidal plane.
        ax
            Matplotlib axes to plot the mesh. If None, a new figure is created.
        cmap
            Colormap to use for the data, by default 'viridis'.

        Returns
        -------
        `~matplotlib.axes.Axes`
            Matplotlib axes with the plotted mesh.

        Raises
        ------
        ValueError
            If the data array does not have the same number of faces as the grid.
        """
        data = np.asarray_chkfinite(data)
        if data.shape[0] != self._num_faces:
            raise ValueError("The data array must have the same number of faces as the grid.")
        data = np.repeat(data, 2)

        cells = self._cells[: self._num_faces]
        tri = np.empty((self._num_faces * 2, 3), dtype=np.int32)
        for i, cell in enumerate(cells):
            tri[2 * i] = [cell[0], cell[1], cell[2]]
            tri[2 * i + 1] = [cell[0], cell[2], cell[3]]
        verts = self._vertices[: self.num_poloidal, ::2]
        triangles = Triangulation(verts[:, 0], verts[:, 1], tri)

        if ax is None:
            _, ax = plt.subplots(layout="constrained")

        ax.set_aspect(1)
        ax.tripcolor(triangles, data, cmap=cmap)
        ax.set_xlim(self._mesh_extent["rmin"], self._mesh_extent["rmax"])
        ax.set_ylim(self._mesh_extent["zmin"], self._mesh_extent["zmax"])

        if self._coordinate_system == "cartesian":
            ax.set_xlabel("X [m]")
            ax.set_ylabel("Y [m]")
        elif self._coordinate_system == "cylindrical":
            ax.set_xlabel("R [m]")
            ax.set_ylabel("Z [m]")

        return ax



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for common utilities to load profiles."""

from dataclasses import dataclass
from typing import Literal

from raysect.core.math import Vector3D
from raysect.core.math.function.float import Function2D, Function3D
from raysect.core.math.function.vector3d import Constant3D as ConstantVector3D
from raysect.core.math.function.vector3d import Function2D as VectorFunction2D
from raysect.core.math.function.vector3d import Function3D as VectorFunction3D

from ..ids.common.species import SpeciesComposition

__all__ = [
    "ProfileInterpolator",
    "warn_unsupported_species",
    "get_subset_name_index",
]


ZERO_VELOCITY = ConstantVector3D(Vector3D(0, 0, 0))


@dataclass
class ProfileInterpolator:
    """Dataclass to hold the interpolators for profiles."""

    density: Function2D | Function3D | None = None
    """Interpolating function for the density profile. Prefer ``density_thermal`` if available."""
    density_thermal: Function2D | Function3D | None = None
    """Interpolating function for the thermal density profile."""
    density_fast: Function2D | Function3D | None = None
    """Interpolating function for the fast density profile."""
    temperature: Function2D | Function3D | None = None
    """Interpolating function for the temperature profile."""
    velocity: VectorFunction2D | VectorFunction3D | None = None
    """Interpolating function for the velocity profile."""


def warn_unsupported_species(
    composition: SpeciesComposition,
    species_type: Literal["ion_bundle", "molecule", "molecular_bundle"],
) -> None:
    """Warn if species of a given type are present in the composition dictionary.

    Parameters
    ----------
    composition
        Instance of the `.SpeciesComposition` dataclass
    species_type
        Type of species to check for (e.g., 'ion_bundle', 'molecule', 'molecular_bundle').
    """
    profiles = getattr(composition, species_type, None)
    if profiles is not None and len(profiles) > 0:
        names: list[str] = []
        for profile_data in profiles:
            name: str | None = getattr(profile_data.species, "name", None)
            if name is None:
                element = getattr(profile_data.species, "element", None)
                if element is None:
                    elements = getattr(profile_data.species, "elements", None)
                    if elements is None or len(elements) == 0:
                        name = "Unknown"
                    else:
                        name = "".join([e.name for e in elements])
                else:
                    name = element.name

            names.append(name)

        print(
            f"Warning! Species of type '{species_type}' are currently not supported.\n"
            + f"The following species will be skipped: {'; '.join(names)}"
        )


def get_subset_name_index(subset_id_dict: dict, grid_subset_id: int | str) -> tuple[str, int]:
    """Get the name and index of a grid subset from its identifier.

    Parameters
    ----------
    subset_id_dict
        Dictionary with grid subset indices.
    grid_subset_id
        Identifier of the grid subset. Either index or name.

    Returns
    -------
    grid_subset_name
        Name of the grid subset.
    grid_subset_index
        Index of the grid subset.

    Raises
    ------
    ValueError
        If the grid subset with the given identifier is not found.
    """
    subset_id = subset_id_dict.copy()
    subset_id.update({value: key for key, value in subset_id.items()})

    try:
        grid_subset_index = int(grid_subset_id)
        grid_subset_name = subset_id[grid_subset_index]
    except ValueError as err1:
        try:
            grid_subset_name = str(grid_subset_id)
            grid_subset_index = subset_id[grid_subset_name]
        except KeyError:
            raise ValueError(f"Unable to find a grid subset with ID {grid_subset_id}.") from err1
    except KeyError as err2:
        raise ValueError(f"Unable to find a grid subset with ID {grid_subset_id}.") from err2

    return grid_subset_name, grid_subset_index



import pytest

from cherab.imas.datasets._fetchers import fetch_data
from cherab.imas.datasets._registry import method_files_map


def test_fetchers_missing_pooch():
    with pytest.raises(ImportError):
        _ = fetch_data("sample_data", data_fetcher=None)


def test_fetchers_without_progress():
    data_path = fetch_data(method_files_map["iter_jintrac"][0], show_progress=False)
    assert method_files_map["iter_jintrac"][0] in data_path


def test_fetchers_error_handling():
    with pytest.raises(RuntimeError):
        _ = fetch_data("non_existent_dataset")



"""Provide functionality to fetch IMAS sample datasets."""

from pathlib import Path

from ...imas import __version__
from ._patch import fix_jintrac
from ._registry import registry

try:
    import pooch

except ImportError:
    pooch = None
    data_fetcher = None
else:
    data_fetcher = pooch.create(
        path=pooch.os_cache("cherab/imas"),
        base_url="doi:10.5281/zenodo.17062699",
        registry=registry,
    )


def fetch_data(dataset_name: str, data_fetcher=data_fetcher, show_progress: bool = True) -> str:
    if data_fetcher is None or pooch is None:
        raise ImportError(
            "Missing optional dependency 'pooch' required for cherab.imas.datasets module. "
            "Please use pip or conda to install 'pooch'."
        )
    if show_progress:
        from ._progress import PoochRichProgress

        progress = PoochRichProgress(filename=dataset_name)
    else:
        progress = None

    downloader = pooch.DOIDownloader(
        headers={"User-Agent": f"CHERAB-IMAS {__version__}"}, progressbar=progress, timeout=60
    )

    #  The "fetch" method returns the full path to the downloaded data file.
    try:
        return data_fetcher.fetch(
            dataset_name,
            downloader=downloader,  # pyrefly: ignore[bad-argument-type]
        )
    except Exception as e:
        raise RuntimeError(f"Failed to fetch dataset '{dataset_name}'.") from e
    finally:
        # Ensure progress bar is closed properly
        if progress is not None:
            progress.close()


def iter_jintrac() -> str:
    """Fetch and return the path to the ITER JINTRAC sample dataset.

    Returns
    -------
    str
        Path to the downloaded ITER JINTRAC sample dataset file.

    Examples
    --------
    >>> from cherab.imas import datasets
    >>> data_path = datasets.iter_jintrac()
    >>> data_path
    '.../cherab/imas/iter_scenario_53298_seq1_DD4_mod.nc'
    """
    path = fetch_data("iter_scenario_53298_seq1_DD4.nc")

    # NOTE: Apply patch to fix datasets
    path_in = Path(path)
    path_out = (path_in.parent / f"{path_in.stem}_mod").with_suffix(path_in.suffix)
    if not path_out.exists():
        fix_jintrac(path, str(path_out))
    path = str(path_out)

    return path


def iter_solps() -> str:
    """Fetch and return the path to the ITER SOLPS sample dataset.

    Returns
    -------
    str
        Path to the downloaded ITER SOLPS sample dataset file.

    Examples
    --------
    >>> from cherab.imas import datasets
    >>> data_path = datasets.iter_solps()
    >>> data_path
    '.../cherab/imas/iter_scenario_123364_1.nc'
    """
    return fetch_data("iter_scenario_123364_1.nc")


def iter_jorek() -> str:
    """Fetch and return the path to the ITER JOREK sample dataset.

    Returns
    -------
    str
        Path to the downloaded ITER JOREK sample dataset file.

    Examples
    --------
    >>> from cherab.imas import datasets
    >>> data_path = datasets.iter_jorek()
    >>> data_path
    '.../cherab/imas/iter_disruption_113112_1.nc'
    """
    return fetch_data("iter_disruption_113112_1.nc")



from numpy import float64, int32
from numpy.typing import NDArray

def cell_to_5tetra(cells: NDArray[int32]) -> NDArray[int32]:
    """Generate tetrahedral indices by dividing one cell into 5 tetrahedra.

    One cubic-like cell having 8 vertices can be divided into a minimum of five tetrahedra.
    Reference: https://www.mathworks.com/matlabcentral/mlc-downloads/downloads/submissions/48509/versions/3/previews/COMP_GEOM_TLBX/html/Divide_hypercube_5_simplices_3D.html

    .. note::
        If one side face of the cell is twisted, Adjacent cells do not share a plane with each
        other.
        In this case, it is required to alternate combinations of tetrahedra, or simply use
        :func:`~cherab.imas.math.tetrahedralize.cell_to_6tetra`.

    Parameters
    ----------
    cells : (N,8) ndarray [numpy.int32]
        cell indices 2D array, the shape of which is :math:`(N, 8)`, where :math:`N` is the number
        of cells.

    Returns
    -------
    `(5N,4) ndarray`
        tetrahedra indices array, the shape of which is :math:`(5N, 4)`.

    Examples
    --------
    >>> import numpy as np
    >>> from cherab.imas.math.tetrahedralize import cell_to_5tetra
    >>>
    >>> array = np.arange(16, dtype=np.int32).reshape((2, -1))
    >>> array
    array([[ 0,  1,  2,  3,  4,  5,  6,  7],
           [ 8,  9, 10, 11, 12, 13, 14, 15]], dtype=int32)
    >>> cell_to_5tetra(array)
        array([[ 0,  1,  3,  4],
               [ 1,  3,  4,  6],
               [ 3,  6,  7,  4],
               [ 1,  2,  3,  6],
               [ 1,  6,  4,  5],
               [ 8,  9, 11, 12],
               [ 9, 11, 12, 14],
               [11, 14, 15, 12],
               [ 9, 10, 11, 14],
               [ 9, 14, 12, 13]], dtype=int32)
    """

def cell_to_6tetra(cells: NDArray[int32]) -> NDArray[int32]:
    """Generate tetrahedral indices by dividing one cell into 6 tetrahedra.

    One cubic-like cell having 8 vertices can be divided into six tetrahedra.
    This manner is useful when the cell is twisted.

    Parameters
    ----------
    cells : (N,8) ndarray [numpy.int32]
        Cell indices 2D array, the shape of which is :math:`(N, 8)`, where :math:`N` is the number
        of cells.

    Returns
    -------
    `(6N,4) ndarray`
        Tetrahedra indices array, the shape of which is :math:`(6N, 4)`.

    Examples
    --------
    >>> import numpy as np
    >>> from cherab.imas.math.tetrahedralize import cell_to_6tetra
    >>>
    >>> array = np.arange(16, dtype=np.int32).reshape((2, -1))
    >>> array
    array([[ 0,  1,  2,  3,  4,  5,  6,  7],
           [ 8,  9, 10, 11, 12, 13, 14, 15]], dtype=int32)
    >>> cell_to_6tetra(array)
    array([[ 6,  2,  1,  0],
           [ 7,  3,  2,  0],
           [ 0,  7,  6,  2],
           [ 1,  5,  6,  4],
           [ 0,  4,  6,  7],
           [ 6,  4,  0,  1],
           [14, 10,  9,  8],
           [15, 11, 10,  8],
           [ 8, 15, 14, 10],
           [ 9, 13, 14, 12],
           [ 8, 12, 14, 15],
           [14, 12,  8,  9]], dtype=int32)
    """

def calculate_tetra_volume(vertices: NDArray[float64], tetrahedra: NDArray[int32]) -> float:
    """Calculate the volume of tetrahedra.

    Parameters
    ----------
    vertices : ndarray [numpy.float64]
        Vertices of tetrahedra.
    tets : ndarray [numpy.int32]
        Tetrahedra indices.

    Returns
    -------
    float
        Volume of tetrahedra.

    Examples
    --------
    >>> vertices = np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])
    >>> tetras = np.array([[0, 1, 2, 3]], dtype=np.int32)
    >>> calculate_tetra_volume(vertices, tetras)
    0.16666666666666666
    """



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Subpackage for creating plasma/equilibrium/magnetic field objects from IMAS."""

from .blend import load_plasma
from .core import load_core_plasma
from .edge import load_edge_plasma
from .equilibrium import load_equilibrium, load_magnetic_field

__all__ = [
    "load_plasma",
    "load_core_plasma",
    "load_edge_plasma",
    "load_equilibrium",
    "load_magnetic_field",
]



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for loading 2D plasma equilibrium data from the equilibrium IDS."""

from __future__ import annotations

from dataclasses import dataclass

import numpy as np
from numpy.typing import NDArray
from raysect.core.math import Point2D

from imas.ids_defs import EMPTY_FLOAT, EMPTY_INT
from imas.ids_toplevel import IDSToplevel

RECTANGULAR_GRID = 1

__all__ = ["load_equilibrium_data"]


@dataclass
class Equilibrium2DData:
    """Dataclass to hold 2D plasma equilibrium data."""

    r: NDArray[np.float64]
    z: NDArray[np.float64]
    psi_grid: NDArray[np.float64]
    psi_axis: float
    psi_lcfs: float
    magnetic_axis: Point2D
    x_points: list[Point2D]
    strike_points: list[Point2D]
    psi_norm: NDArray[np.float64]
    f: NDArray[np.float64]
    q: NDArray[np.float64]
    phi: NDArray[np.float64] | None
    rho_tor_norm: NDArray[np.float64] | None
    b_vacuum_radius: float
    b_vacuum_magnitude: float
    lcfs_polygon: NDArray[np.float64]
    time: float


def load_equilibrium_data(
    equilibrium_ids: IDSToplevel,
) -> Equilibrium2DData:
    """Load 2D plasma equilibrium data from the equilibrium IDS.

    Parameters
    ----------
    equilibrium_ids
        The time-slice of the equilibrium IDS.

    Returns
    -------
    Equilibrium2DData
        Dataclass instance containing the following attributes:

        :r: ``(N, )`` ndarray with R coordinates of rectangular grid,
        :z: ``(M, )`` ndarray with Z coordinates of rectangular grid,
        :psi_grid: ``(N, M)`` ndarray with psi grid values,
        :psi_axis: The psi value at the magnetic axis,
        :psi_lcfs: The psi value at the LCFS,
        :magnetic_axis: Point2D containing the coordinates of the magnetic axis,
        :x_points: list or tuple of Point2D x-points,
        :strike_points: list or tuple of Point2D strike-points,
        :psi_norm: ``(K, )`` ndarray with the values of psi_norm,
        :f: ``(K, )`` ndarray with the current flux profile on psi_norm,
        :q: ``(K, )`` ndarray with the safety factor (q) profile on psi_norm,
        :phi: ``(K, )`` ndarray with the toroidal flux profile on psi_norm,
        :rho_tor_norm: ``(K, )`` ndarray with the normalised toroidal flux coordinate on psi_norm,
        :b_vacuum_radius: Vacuum B-field reference radius (in meters),
        :b_vacuum_magnitude: Vacuum B-Field magnitude at the reference radius,
        :lcfs_polygon: ``(2, L)`` ndarray of ``[[x0, ...], [y0, ...]]`` vertices specifying the LCFS
            boundary
        :time: The time stamp of the time-slice (in seconds).

    Raises
    ------
    RuntimeError
        If the required data is not available in the IDS.
    """
    if not len(equilibrium_ids.time_slice):
        raise RuntimeError("Equilibrium IDS does not have a time slice.")

    profiles_2d = equilibrium_ids.time_slice[0].profiles_2d

    rectangular_grid = False
    for prof2d in profiles_2d:
        if prof2d.grid_type.index == RECTANGULAR_GRID or prof2d.grid_type.index == EMPTY_INT:
            rectangular_grid = True
            break

    if not rectangular_grid:
        raise RuntimeError(
            "Unable to read equilibrium: rectangular grid for psi profile is not found and other grid types are not supported."
        )

    r = np.asarray_chkfinite(prof2d.grid.dim1)
    z = np.asarray_chkfinite(prof2d.grid.dim2)

    psi_grid = np.asarray_chkfinite(prof2d.psi)
    if psi_grid.shape != (r.size, z.size):
        raise RuntimeError(
            "Unable to read equilibrium: the shape of 'profiles_2d[i].psi' does not match the grid shape."
        )

    profiles_1d = equilibrium_ids.time_slice[0].profiles_1d

    psi1d = np.asarray_chkfinite(profiles_1d.psi)
    if not psi1d.size:
        raise RuntimeError("Unable to read equilibrium: 'profiles_1d.psi' is empty.")

    if not len(profiles_1d.f):
        raise RuntimeError("Unable to read equilibrium: 'profiles_1d.f' is empty.")

    if not len(profiles_1d.q):
        raise RuntimeError("Unable to read equilibrium: 'profiles_1d.q' is empty.")

    global_quantities = equilibrium_ids.time_slice[0].global_quantities

    psi_axis = global_quantities.psi_axis
    if psi_axis == EMPTY_FLOAT:
        raise RuntimeError("Unable to read equilibrium: 'global_quantities.psi_axis' is not set.")
    else:
        psi_axis = float(psi_axis)

    psi_lcfs = global_quantities.psi_boundary
    if psi_lcfs == EMPTY_FLOAT:
        raise RuntimeError(
            "Unable to read equilibrium: 'global_quantities.psi_boundary' is not set."
        )
    else:
        psi_lcfs = float(psi_lcfs)

    psi_norm = (psi1d - psi_axis) / (psi_lcfs - psi_axis)
    psi_norm, index = np.unique(psi_norm, return_index=True)
    f: NDArray[np.float64] = profiles_1d.f[index]
    q: NDArray[np.float64] = profiles_1d.q[index]

    # additional 1D profiles
    phi: NDArray[np.float64] | None = profiles_1d.phi[index] if len(profiles_1d.phi) else None
    rho_tor_norm: NDArray[np.float64] | None = (
        profiles_1d.rho_tor_norm[index] if len(profiles_1d.rho_tor_norm) else None
    )
    r_axis = global_quantities.magnetic_axis.r
    z_axis = global_quantities.magnetic_axis.z
    if r_axis == EMPTY_FLOAT or z_axis == EMPTY_FLOAT:
        raise RuntimeError("Unable to read equilibrium: magnetic axis is not set.")

    magnetic_axis = Point2D(r_axis.value, z_axis.value)

    boundary = equilibrium_ids.time_slice[0].boundary

    x_points = []
    if hasattr(boundary, "x_point"):  # DD4 no longer has x_point
        for x_point in boundary.x_point:
            if x_point.r != EMPTY_FLOAT and x_point.z != EMPTY_FLOAT:
                x_points.append(Point2D(x_point.r.value, x_point.z.value))

    strike_points = []
    if hasattr(boundary, "strike_point"):  # DD4 no longer has strike_point
        for strike_point in boundary.strike_point:
            if strike_point.r != EMPTY_FLOAT and strike_point.z != EMPTY_FLOAT:
                strike_points.append(Point2D(strike_point.r.value, strike_point.z.value))

    r_lcfs = boundary.outline.r
    z_lcfs = boundary.outline.z

    lcfs_polygon = np.array([r_lcfs, z_lcfs])

    # exclude the recurring points from the polygon
    if lcfs_polygon.shape[1]:
        _, index = np.unique(lcfs_polygon, return_index=True, axis=1)
        lcfs_polygon = lcfs_polygon[:, np.sort(index)]

    if lcfs_polygon.shape[1] < 3:
        raise RuntimeError(
            "Unable to read equilibrium: boundary.outline contains less than 3 unique points."
        )

    b_vacuum_radius = equilibrium_ids.vacuum_toroidal_field.r0
    if b_vacuum_radius == EMPTY_FLOAT:
        raise RuntimeError("Unable to read equilibrium: vacuum_toroidal_field.r0 is not set.")
    else:
        b_vacuum_radius = float(b_vacuum_radius)

    b_vacuum_magnitude = equilibrium_ids.vacuum_toroidal_field.b0[0]
    if b_vacuum_magnitude == EMPTY_FLOAT:
        raise RuntimeError("Unable to read equilibrium: vacuum_toroidal_field.b0 is not set.")
    else:
        b_vacuum_magnitude = float(b_vacuum_magnitude)

    time = float(equilibrium_ids.time[0])

    return Equilibrium2DData(
        r=r,
        z=z,
        psi_grid=psi_grid,
        psi_axis=psi_axis,
        psi_lcfs=psi_lcfs,
        magnetic_axis=magnetic_axis,
        x_points=x_points,
        strike_points=strike_points,
        psi_norm=psi_norm,
        f=f,
        q=q,
        phi=phi,
        rho_tor_norm=rho_tor_norm,
        b_vacuum_radius=b_vacuum_radius,
        b_vacuum_magnitude=b_vacuum_magnitude,
        lcfs_polygon=lcfs_polygon,
        time=time,
    )



"""The configuration for this package documentation."""

from datetime import date

from packaging.version import parse

from cherab.imas import __version__

# -- Project information -----------------------------------------------------
project = "CHERAB-IMAS"
author = "CHERAB Team"
copyright = f"2023-{date.today().year}, {author}"
version_obj = parse(__version__)
release = version_obj.public

# -- General configuration ---------------------------------------------------
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.todo",
    "sphinx.ext.napoleon",
    "sphinx_api_relink",
    "sphinx_copybutton",
    "sphinx_codeautolink",
    "sphinx_design",
    "sphinx_github_style",
    "sphinx_immaterial",
    "IPython.sphinxext.ipython_console_highlighting",
    "myst_parser",
    "nbsphinx",
]

default_role = "obj"

# autodoc config
autodoc_member_order = "bysource"
autodoc_typehints = "signature"
autodoc_type_aliases = {
    "ArrayLike": "ArrayLike",
}

# autosummary config
autosummary_generate = True
autosummary_generate_overwrite = True
autosummary_imported_members = True
autosummary_ignore_module_all = False

# napoleon config
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_attr_annotations = True
napoleon_use_rtype = False

# todo config
todo_include_todos = True

# Strip prompt text when copying code blocks with copy button
copybutton_exclude = ".linenos, .gp"
copybutton_prompt_text = r">>> |\.\.\. |\$ "
copybutton_prompt_is_regexp = True

exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "**.ipynb_checkpoints",
]

templates_path = ["_templates"]

# myst-parser config
myst_enable_extensions = [
    "amsmath",
    "attrs_inline",
    "colon_fence",
    "deflist",
    "dollarmath",
    "fieldlist",
    "html_admonition",
    "html_image",
    "linkify",
    "replacements",
    "smartquotes",
    "strikethrough",
    "substitution",
    "tasklist",
]
myst_url_schemes = {
    "http": None,
    "https": None,
    "mailto": None,
    "ftp": None,
    "wiki": "https://en.wikipedia.org/wiki/{{path}}#{{fragment}}",
    "doi": "https://doi.org/{{path}}",
}

# -- HTML output ------------------------------------------------------------
html_theme = "sphinx_immaterial"
html_title = project
html_theme_options = {
    "repo_url": "https://github.com/cherab/imas",
    "repo_name": "CHERAB-IMAS",
    "edit_uri": "blob/master/docs/source",
    "icon": {
        "repo": "fontawesome/brands/github",
    },
    "features": [
        "navigation.expand",
        # "navigation.tabs",
        # "navigation.tabs.sticky",
        # "toc.integrate",
        "navigation.sections",
        # "navigation.instant",
        # "header.autohide",
        "navigation.top",
        "navigation.footer",
        # "navigation.tracking",
        # "search.highlight",
        "search.share",
        "search.suggest",
        "toc.follow",
        "toc.sticky",
        "content.tabs.link",
        # "content.code.copy",
        # "content.action.edit",
        # "content.action.view",
        "content.tooltips",
        "announce.dismiss",
    ],
    # "toc_title_is_page_title": True,
    # "globaltoc_collapse": True,
    "palette": [
        {
            "media": "(prefers-color-scheme: light)",
            "scheme": "default",
            "primary": "indigo",
            "toggle": {
                "icon": "material/lightbulb-outline",
                "name": "Switch to dark mode",
            },
        },
        {
            "media": "(prefers-color-scheme: dark)",
            "scheme": "slate",
            "primary": "indigo",
            "toggle": {
                "icon": "material/lightbulb",
                "name": "Switch to light mode",
            },
        },
    ],
}
html_static_path = ["_static"]
html_css_files = ["custom.css"]

# Shorten Table Of Contents in API documentation
object_description_options = [
    (".*", dict(include_fields_in_toc=False)),
    (".*parameter", dict(include_in_toc=False)),
]

# -- Intersphinx configuration -----------------------------------------------
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
    "raysect": ("http://www.raysect.org", None),
    "cherab": ("https://www.cherab.info", None),
    "imas-python": ("https://imas-python.readthedocs.io/en/stable/", None),
    "rich": ("https://rich.readthedocs.io/en/stable/", None),
    "pooch": ("https://www.fatiando.org/pooch/latest/", None),
    "ultraplot": ("https://ultraplot.readthedocs.io/en/stable/", None),
}

intersphinx_timeout = 10

# -- Sphinx GitHub Style configuration ----------------------------------------
linkcode_blob = "master" if version_obj.is_devrelease else f"v{version_obj.public}"
linkcode_url = "https://github.com/cherab/imas"
linkcode_link_text = "Source"

# -- NBSphinx configuration ---------------------------------------------------
# nbsphinx_execute = "never"
nbsphinx_thumbnails = {
    "notebooks/observer/bolometer": "_static/images/bolometer.png",
}
nbsphinx_prolog = r"""
{% set docname = 'docs/' + env.doc2path(env.docname, base=None)|string %}

.. raw:: html

    <div class="admonition note">
      This page was generated from
      <a class="reference external" href="https://github.com/cherab/imas/blob/{{ env.config.linkcode_blob|e }}/{{ docname|e }}">{{ docname|e }}</a>.
      <br />
      <a href="{{ env.docname.split('/')|last|e + '.ipynb' }}" class="reference download internal" download>Download notebook</a>.
      <script>
        if (document.location.host) {
          let nbviewer_link = document.createElement('a');
          nbviewer_link.setAttribute('href',
            'https://nbviewer.org/url' +
            (window.location.protocol == 'https:' ? 's/' : '/') +
            window.location.host +
            window.location.pathname.slice(0, -4) +
            'ipynb');
          nbviewer_link.innerHTML = 'Or view it on <em>nbviewer</em>';
          nbviewer_link.classList.add('reference');
          nbviewer_link.classList.add('external');
          document.currentScript.replaceWith(nbviewer_link, '.');
        }
      </script>
    </div>

.. raw:: latex

    \nbsphinxstartnotebook{\scriptsize\noindent\strut
    \textcolor{gray}{The following section was generated from
    \sphinxcode{\sphinxupquote{\strut {{ docname | escape_latex }}}} \dotfill}}
"""



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for loading unstructured 2D grids from IMAS grid_ggd IDS structure."""

from enum import IntEnum
from typing import Literal, overload

import numpy as np
from numpy.typing import NDArray

from imas.ids_defs import EMPTY_INT
from imas.ids_structure import IDSStructure

from ....ggd.unstruct_2d_mesh import UnstructGrid2D

__all__ = ["load_unstruct_grid_2d"]


class DIMENSION(IntEnum):
    """Enumeration for grid dimensions."""

    VERTEX = 0
    EDGE = 1
    FACE = 2


@overload
def load_unstruct_grid_2d(
    grid_ggd: IDSStructure, space_index: int = 0, with_subsets: Literal[False] = False
) -> UnstructGrid2D: ...


@overload
def load_unstruct_grid_2d(
    grid_ggd: IDSStructure,
    space_index: int = 0,
    *,
    with_subsets: Literal[True],
) -> tuple[UnstructGrid2D, dict[str, NDArray[np.int32]], dict[str, int]]: ...


def load_unstruct_grid_2d(
    grid_ggd: IDSStructure, space_index: int = 0, with_subsets: bool = False
) -> UnstructGrid2D | tuple[UnstructGrid2D, dict[str, NDArray[np.int32]], dict[str, int]]:
    """Load unstructured 2D grid from the grid_ggd structure.

    Parameters
    ----------
    grid_ggd
        The grid_ggd structure.
    space_index
        The index of the grid space, by default 0.
    with_subsets
        Read grid subset data, by default False.

    Returns
    -------
    grid : `.UnstructGrid2D`
        Unstructured 2D grid object.
    subsets : `dict[str, NDArray[numpy.int32]]`
        Dictionary with grid subsets for each subset name containing the indices of the cells from
        that subset. Note that 'Cells' subset is included only if cell indices are specified.
    subset_id : `dict[str, int]`
        Dictionary with grid subset indices.

    Raises
    ------
    ValueError
        If the specified space does not contain a 2D grid.
    """
    space = grid_ggd.space[space_index]

    # Check if the grid is 2D
    if len(space.objects_per_dimension) != 3:
        raise ValueError("The load_unstruct_grid_2d() supports only unstructured 2D grids.")

    grid_name = str(grid_ggd.identifier.name)

    # Reading vertices
    num_vert = len(space.objects_per_dimension[DIMENSION.VERTEX].object)
    vertices = np.empty((num_vert, 2), dtype=np.float64)
    for i in range(num_vert):
        vertices[i] = space.objects_per_dimension[DIMENSION.VERTEX].object[i].geometry[:2]

    # Reading polygonal cells
    cells = []
    winding_ok = True
    for object in space.objects_per_dimension[DIMENSION.FACE].object:
        # getting cell from nodes
        cell = np.asarray_chkfinite(object.nodes, dtype=np.int32) - 1  # Fortran to C indexing
        if cell.size > 3:
            # trying to get the nodes in winding order by parsing the edges
            edge_dict = {}
            for boundary in object.boundary:
                n1, n2 = (
                    space.objects_per_dimension[DIMENSION.EDGE].object[boundary.index - 1].nodes - 1
                )  # Fortran to C indexing
                if n1 not in cell or n2 not in cell:  # fail, error in the data
                    edge_dict = {}
                    break
                if n1 in edge_dict:
                    if n2 == edge_dict[n1][0]:  # fail, error in the data
                        edge_dict = {}
                        break
                    edge_dict[n1][1] = n2
                else:
                    edge_dict[n1] = [n2, -1]
                if n2 in edge_dict:
                    if n1 == edge_dict[n2][0]:  # fail, error in the data
                        edge_dict = {}
                        break
                    edge_dict[n2][1] = n1
                else:
                    edge_dict[n2] = [n1, -1]
            if len(edge_dict) == cell.size:  # success, getting the cell nodes in winding order
                cell1 = np.empty(len(edge_dict), dtype=np.int32)
                cell1[0] = cell[0]
                pair = edge_dict[cell1[0]]
                cell1[1] = cell[1] if cell[1] in pair else pair[0]
                for i in range(2, cell1.size):
                    pair = edge_dict[cell1[i - 1]]
                    cell1[i] = pair[1] if cell1[i - 2] == pair[0] else pair[0]
                cell = cell1
            else:
                winding_ok = False

            cells.append(cell)

    if not winding_ok:
        print("Warning! Unable to verify that the cell nodes are in the winding order.")

    grid = UnstructGrid2D(vertices, cells, name=grid_name)

    if not with_subsets:
        return grid

    # Reading grid subsets (2D only)
    cell_subset_ids = (5, 22, 23, 24, 25, 38, 39, 40)
    subsets = {}
    subset_id = {}
    for subset in grid_ggd.grid_subset:
        dimension_is_2d = subset.dimension == DIMENSION.FACE + 1  # C to Fortran indexing
        known_subset_id = (
            subset.dimension != EMPTY_INT and subset.identifier.index in cell_subset_ids
        )
        if (dimension_is_2d or known_subset_id) and len(subset.element):
            name = str(subset.identifier.name)
            indices = np.empty(len(subset.element), dtype=np.int32)
            for i, element in enumerate(subset.element):
                if len(element.object) > 1:
                    print(
                        f"Warning! Skipping grid subset {name}, "
                        + "because it includes cells not present in the original grid."
                    )
                    break
                indices[i] = element.object[0].index.value
            subsets[name] = indices - 1  # Fortran to C indexing
            subset_id[name] = subset.identifier.index.value

    return grid, subsets, subset_id



#!/usr/bin/env -S pixi run python
"""Calculate camera images with Full plasma profile.

This demo loads plasma from the DINA-JINTRAC scenario frame by frame, samples 4 deuterium and 4
beryllium line emission profiles and convolves them with the geometry matrices for the 55.G1 or
55.E2 cameras.

Then converts spectral data to RGB images.
"""

from multiprocessing import get_context
from os import system
from pathlib import Path

import numpy as np
from imas import DBEntry
from imas.ids_defs import HDF5_BACKEND, MDSPLUS_BACKEND
from matplotlib import pyplot as plt
from raysect.core.math import Point3D, Vector3D
from raysect.optical import Spectrum
from raysect.optical.colour import ciexyz_to_srgb, resample_ciexyz, spectrum_to_ciexyz
from scipy.interpolate import RectBivariateSpline, griddata
from scipy.sparse import coo_matrix

from cherab.core.atomic import Line, beryllium, deuterium
from cherab.core.model import ExcitationLine, RecombinationLine
from cherab.imas.plasma import load_plasma
from cherab.openadas import OpenADAS

NPROC = 8  # number of parallel processes

DATABASE = "ITER"
DATABASE_MD = "ITER_MD"
USER = "public"
USER_MD = "public"

SHOT, RUN = 134174, 117
time = np.arange(10.5, 75.1, 0.5)

# 55.G1
geom_matrix_shot = 150701

# EP03 Divertor
# geom_matrix_run = 1000
# EP03 Left
# geom_matrix_run = 1010
# EP03 Right
geom_matrix_run = 1020
# EP03 Upper
# geom_matrix_run = 1030

# 55.E2
# geom_matrix_shot = 150502

# EP11 Lower (A0)
# geom_matrix_run = 1000
# EP11 Upper (B0)
# geom_matrix_run = 1010
# EP12 (C0)
# geom_matrix_run = 1020

unsaturated_fraction = 0.95


def read_geometry_matrix_without_reflections(  # noqa: D103
    shot, run, user, database="ITER_MD", backend=HDF5_BACKEND
):
    geometry_matrix = {}

    with DBEntry(backend, database, shot, run, user) as entry:
        camera = entry.get("camera_visible")

    geometry_matrix["camera_name"] = camera.channel[0].name
    geometry_matrix["pixel_to_alpha"] = camera.channel[0].detector[0].pixel_to_alpha
    geometry_matrix["pixel_to_beta"] = camera.channel[0].detector[0].pixel_to_beta

    geometry_matrix["voxel_map"] = camera.channel[0].detector[0].geometry_matrix.voxel_map

    emission_grid = {}
    emission_grid["r"] = camera.channel[0].detector[0].geometry_matrix.emission_grid.dim1
    emission_grid["z"] = camera.channel[0].detector[0].geometry_matrix.emission_grid.dim2
    geometry_matrix["emission_grid"] = emission_grid

    without_reflections = {}
    without_reflections["data"] = (
        camera.channel[0].detector[0].geometry_matrix.without_reflections.data
    )
    without_reflections["voxel_indices"] = (
        camera.channel[0].detector[0].geometry_matrix.without_reflections.voxel_indices
    )
    without_reflections["pixel_indices"] = (
        camera.channel[0].detector[0].geometry_matrix.without_reflections.pixel_indices
    )
    geometry_matrix["without_reflections"] = without_reflections

    return geometry_matrix


def read_geometry_matrix_interpolated(shot, run, user, database="ITER_MD", backend=HDF5_BACKEND):  # noqa: D103
    geometry_matrix = {}

    with DBEntry(backend, database, shot, run, user) as entry:
        camera = entry.get("camera_visible")

    geometry_matrix["camera_name"] = camera.channel[0].name
    geometry_matrix["pixel_to_alpha"] = camera.channel[0].detector[0].pixel_to_alpha
    geometry_matrix["pixel_to_beta"] = camera.channel[0].detector[0].pixel_to_beta

    interpolated = {}
    interpolated["r"] = camera.channel[0].detector[0].geometry_matrix.interpolated.r
    interpolated["z"] = camera.channel[0].detector[0].geometry_matrix.interpolated.z
    interpolated["data"] = camera.channel[0].detector[0].geometry_matrix.interpolated.data
    geometry_matrix["interpolated"] = interpolated

    return geometry_matrix


# Define spectral lines
dalpha = Line(deuterium, 0, (3, 2))
dbeta = Line(deuterium, 0, (4, 2))
dgamma = Line(deuterium, 0, (5, 2))
ddelta = Line(deuterium, 0, (6, 2))
be0_457nm = Line(beryllium, 0, ("2s1 3d1 1d2.0", "2s1 2p1 1p1.0"))
be1_436nm = Line(beryllium, 1, ("4d1 2d4.5", "3p1 2p2.5"))
be1_467nm = Line(beryllium, 1, ("4f1 2f6.5", "3d1 2d4.5"))
be1_527nm = Line(beryllium, 1, ("4s1 2s0.5", "3p1 2p2.5"))

spectral_lines = [dalpha, dbeta, dgamma, ddelta, be0_457nm, be1_436nm, be1_467nm, be1_527nm]

# Sample emission profiles

# sampling points
r = np.linspace(4.0, 8.5, 451)
z = np.linspace(-4.5, 4.6, 911)
volume = np.pi * (r[1:] ** 2 - r[:-1] ** 2)[:, None] * (z[1:] - z[:-1])[None, :]
r = 0.5 * (r[1:] + r[:-1])
z = 0.5 * (z[1:] + z[:-1])
r2d, z2d = np.meshgrid(r, z, indexing="ij")

atomic_data = OpenADAS(permit_extrapolation=True)

line_emission = np.zeros((len(spectral_lines), time.size, r.size, z.size))
line_wavelength = np.array(
    [atomic_data.wavelength(line.element, line.charge, line.transition) for line in spectral_lines]
)

wvl_delta = 3.0
direction = Vector3D(0, 0, 1)
print("Sampling emission profiles...")
for it, t in enumerate(time):
    print(f"Processing time moment: {t} s")
    plasma = load_plasma(MDSPLUS_BACKEND, DATABASE, SHOT, RUN, USER, time=t)
    plasma.atomic_data = atomic_data

    for il, line in enumerate(spectral_lines):
        plasma.models = [ExcitationLine(line), RecombinationLine(line)]
        line_wvl = line_wavelength[il]
        spectrum = Spectrum(line_wvl - 0.5 * wvl_delta, line_wvl + 0.5 * wvl_delta, 1)
        for ir, r1 in enumerate(r):
            for iz, z1 in enumerate(z):
                point = Point3D(r1, 0, z1)
                for model in plasma.models:
                    line_emission[il, it, ir, iz] += model.emission(
                        point, direction, spectrum.new_spectrum()
                    ).total()

# Load geometry matrix without reflections
geom_matrix_norefl = read_geometry_matrix_without_reflections(
    geom_matrix_shot, geom_matrix_run, USER_MD, DATABASE_MD
)
npix_x = geom_matrix_norefl["pixel_to_alpha"].size
npix_y = geom_matrix_norefl["pixel_to_beta"].size
npix = npix_x * npix_y
pixels = geom_matrix_norefl["without_reflections"]["pixel_indices"]
ipix = npix_x * pixels[:, 0] + pixels[:, 1]
jvox = geom_matrix_norefl["without_reflections"]["voxel_indices"]
nvox = geom_matrix_norefl["voxel_map"].max() + 1
geom_matrix = coo_matrix(
    (geom_matrix_norefl["without_reflections"]["data"], (ipix, jvox)), shape=(npix, nvox)
).tocsr()

voxel_map = geom_matrix_norefl["voxel_map"][:, :, 0]
r_gm = geom_matrix_norefl["emission_grid"]["r"]
z_gm = geom_matrix_norefl["emission_grid"]["z"]
ivox, index_v = np.unique(voxel_map, return_index=True)

extent = (
    np.rad2deg(geom_matrix_norefl["pixel_to_alpha"][1]),
    np.rad2deg(geom_matrix_norefl["pixel_to_alpha"][-2]),
    np.rad2deg(geom_matrix_norefl["pixel_to_beta"][-2]),
    np.rad2deg(geom_matrix_norefl["pixel_to_beta"][1]),
)

# Wavelengths of the interpolated geometry matrices
geom_matrix_wvl = np.arange(400.0, 701.0, 50.0)

# Obtain images
images = np.zeros((len(spectral_lines), time.size, npix_y, npix_x))

# Parallel tasks
npix_step = min(4096, npix // NPROC + bool(npix % NPROC))
tasks = [(ibegin, min(ibegin + npix_step, npix)) for ibegin in range(0, npix, npix_step)]

print("Obtaining images for each emission profile...")
gm_index_prev = -1
for il_index, il in enumerate(np.argsort(line_wavelength)):
    # Load interpolated geometry matrix closest to this wavelength
    gm_index = np.argmin(np.abs(geom_matrix_wvl - line_wavelength[il])) + 1
    if gm_index != gm_index_prev:
        gm_interp = read_geometry_matrix_interpolated(
            geom_matrix_shot, geom_matrix_run + gm_index, USER_MD, DATABASE_MD
        )
        node_points = np.array([gm_interp["interpolated"]["r"], gm_interp["interpolated"]["z"]]).T
        interp_data = np.moveaxis(gm_interp["interpolated"]["data"], -1, 0).reshape(
            (node_points.shape[0], npix)
        )
        gm_index_prev = gm_index

    for it, t in enumerate(time):
        print(
            f"Processing time moment: {t} s for emission profile {il_index + 1} out of {len(spectral_lines)}."
        )
        # without reflections
        emission_interp = RectBivariateSpline(r, z, line_emission[il, it], kx=1, ky=1)
        emission_grid = emission_interp(r_gm, z_gm, grid=True)
        emission_prof = (
            emission_grid.flatten()[index_v[1:]]
            if ivox[0] == -1
            else emission_grid.flatten()[index_v]
        )
        images[il, it] = (geom_matrix @ emission_prof).reshape(npix_y, npix_x)

        # add reflections
        index = np.where(line_emission[il, it] > 0)
        power = volume[index] * line_emission[il, it][index]
        points = np.array([r2d[index], z2d[index]]).T

        def _f(task, node_points, interp_data, points, power):
            ibegin = task[0]
            iend = task[1]
            interp_matrix = griddata(
                node_points,
                interp_data[:, ibegin:iend],
                points,
                fill_value=0,
                method="linear",
            )

            return power @ interp_matrix

        ctx = get_context(method="fork")

        with ctx.Pool(processes=NPROC) as pool:
            results = pool.map(
                _f,
                tasks,
                node_points=node_points,
                interp_data=interp_data,
                points=points,
            )

        reflected_image = np.zeros(npix)
        for task, res in zip(tasks, results, strict=False):
            reflected_image[task[0] : task[1]] = res

        images[il, it] += reflected_image.reshape((npix_y, npix_x))

# Convert to CIEXYZ
print("Converting to CIEXYZ...")
delta_wl = 1.0
min_wavelength = line_wavelength.min() - 0.5 * delta_wl
max_wavelength = line_wavelength.max() + 0.5 * delta_wl
bins = int(round((max_wavelength - min_wavelength) / delta_wl))
delta_wl = (max_wavelength - min_wavelength) / bins
bin_index = ((line_wavelength - min_wavelength) / delta_wl).astype(int)
resampled_xyz = resample_ciexyz(min_wavelength, max_wavelength, bins)

images_ciexyz = np.zeros((time.size, npix_y, npix_x, 3))
for it in range(time.size):
    spectral_image = np.zeros((npix_y, npix_x, bins))
    for il in range(len(spectral_lines)):
        spectral_image[:, :, bin_index[il]] += images[il, it] / delta_wl
    for it in range(npix_y):
        for ix in range(npix_x):
            spectrum = Spectrum(min_wavelength, max_wavelength, bins)
            spectrum.samples[:] += spectral_image[it, ix]
            images_ciexyz[it, it, ix] = spectrum_to_ciexyz(spectrum, resampled_xyz)

luminance = np.copy(images_ciexyz[:, :, :, 1]).flatten()
luminance.sort()
peak_luminance = luminance[min(luminance.size - 1, int(luminance.size * unsaturated_fraction))]
images_ciexyz /= peak_luminance

demos_path = Path(__file__).parent
frames_path = demos_path / "frames"
frames_path.mkdir(exist_ok=True)

# Convert to SRGB
print("Converting to SRGB and saving...")
camera_name = geom_matrix_norefl["camera_name"].replace(" ", "_")
fig = plt.figure(figsize=(5.0, 6.0), tight_layout=True)
for it, t in enumerate(time):
    ciexyz = images_ciexyz[it]
    rgb_image = np.zeros(ciexyz.shape)
    for it in range(npix_y):
        for ix in range(npix_x):
            rgb_image[it, ix] = ciexyz_to_srgb(
                ciexyz[it, ix, 0], ciexyz[it, ix, 1], ciexyz[it, ix, 2]
            )

    ax = fig.add_subplot(111)
    im = ax.imshow(rgb_image[1:-1, 1:-1], extent=extent)
    ax.set_xlabel(r"$\alpha$, $\degree$")
    ax.set_ylabel(r"$\beta$, $\degree$")
    ax.set_title(f"{SHOT}/{RUN}: {t:.1f} s")
    fig.savefig(frames_path / f"{camera_name}_rgb_{SHOT}_{RUN}_time_{t:.1f}s.png", dpi=300)
    fig.clear()

system(
    "gm convert -delay 12 -loop 0 "
    + f"{frames_path}/*.png {demos_path}/{camera_name}_rgb_{SHOT}_{RUN}.gif"
)



import pytest
from raysect.core.scenegraph import World
from raysect.primitive import Subtract

from cherab.core import Maxwellian, Plasma
from cherab.core.atomic.elements import neon
from cherab.imas.plasma import load_plasma
from cherab.imas.plasma.equilibrium import load_magnetic_field


def test_load_plasma(path_iter_jintrac: str):
    """Test basic loading of plasma data from an IMAS file."""
    plasma = load_plasma(path_iter_jintrac, "r")

    # Test that a Plasma object is returned
    assert isinstance(plasma, Plasma)

    # Check that geometry is set and has the expected type
    assert plasma.geometry is not None
    assert isinstance(plasma.geometry, Subtract)

    # Check that geometry transform is set
    assert plasma.geometry_transform is not None

    # Check that electron distribution exists
    assert plasma.electron_distribution is not None
    assert isinstance(plasma.electron_distribution, Maxwellian)

    # Check if a certain bundled species is split into individual ion species.
    # Here we check only the Ne species as an example, but this can be extended to other species
    # based on the test dataset.
    assert len(plasma.composition) > 0
    ion_charges = set()
    for species in plasma.composition:
        # Check that we have multiple ionization states for Ne
        if species.element == neon:
            ion_charges.add(species.charge)

    # Expecting all charge states from 0 to Z for Ne
    assert len(ion_charges) == neon.atomic_number + 1


def test_load_plasma_with_time(path_iter_jintrac: str):
    """Test loading plasma with specific time parameter."""
    # Test with default time (should not raise error)
    plasma = load_plasma(path_iter_jintrac, "r", time=0.1)
    assert isinstance(plasma, Plasma)

    with pytest.raises(ValueError):
        load_plasma(path_iter_jintrac, "r", time=-1)


def test_load_plasma_with_occurrence(path_iter_jintrac: str):
    """Test loading plasma with different occurrence values."""
    # Test with default occurrence core
    plasma1 = load_plasma(path_iter_jintrac, "r", occurrence_core=0, split_ion_bundles=False)
    assert isinstance(plasma1, Plasma)

    # Test with default occurrence edge
    plasma2 = load_plasma(path_iter_jintrac, "r", occurrence_edge=0, split_ion_bundles=False)
    assert isinstance(plasma2, Plasma)

    # Test with default occurrence core and edge
    plasma3 = load_plasma(
        path_iter_jintrac, "r", occurrence_core=0, occurrence_edge=0, split_ion_bundles=False
    )
    assert isinstance(plasma3, Plasma)


def test_load_plasma_with_parent(path_iter_jintrac: str):
    """Test loading plasma with a parent scene graph node."""
    world = World()

    plasma = load_plasma(path_iter_jintrac, "r", parent=world, split_ion_bundles=False)
    assert isinstance(plasma, Plasma)
    assert plasma.parent is world


def test_load_plasma_with_magnetic_field(path_iter_jintrac: str):
    """Test loading plasma with external magnetic field."""
    # Load magnetic field separately
    try:
        b_field = load_magnetic_field(path_iter_jintrac, "r")
        plasma = load_plasma(path_iter_jintrac, "r", b_field=b_field, split_ion_bundles=False)
        assert isinstance(plasma, Plasma)
        assert plasma.b_field is not None
    except RuntimeError:
        # If magnetic field loading fails, test should still pass
        # as this is handled gracefully in the function
        pass


def test_load_plasma_time_threshold(path_iter_jintrac: str):
    """Test loading plasma with time threshold parameter."""
    # Test with large time threshold (should work)
    plasma = load_plasma(path_iter_jintrac, "r", time_threshold=500, split_ion_bundles=False)
    assert isinstance(plasma, Plasma)

    # Test with very small time threshold (should raise error)
    with pytest.raises(RuntimeError):
        load_plasma(path_iter_jintrac, "r", time_threshold=0.0, split_ion_bundles=False)


def test_load_plasma_not_split_ion_bundles(path_iter_jintrac: str):
    """Test that loading with split_ion_bundles=False."""
    plasma = load_plasma(path_iter_jintrac, "r", split_ion_bundles=False)

    ion_charges_no_split = set()
    for species in plasma.composition:
        if species.element == neon:
            ion_charges_no_split.add(species.charge)

    # Expecting only the bundled charge state
    assert 0 < len(ion_charges_no_split) < neon.atomic_number + 1



import shutil
from pathlib import Path

import pytest

from cherab.imas.datasets import iter_jintrac, iter_jorek, iter_solps


@pytest.fixture(scope="session")
def path_iter_jintrac(tmp_path_factory) -> str:
    """Fixture to provide the path to a sample JINTRAC IMAS dataset."""
    path = Path(iter_jintrac())
    tmp_path = tmp_path_factory.mktemp("cherab-imas-data")
    shutil.move(path, tmp_path)
    return str(tmp_path / path.name)


@pytest.fixture(scope="session")
def path_iter_solps(tmp_path_factory) -> str:
    """Fixture to provide the path to a sample SOLPS IMAS dataset."""
    path = Path(iter_solps())
    tmp_path = tmp_path_factory.mktemp("cherab-imas-data")
    shutil.move(path, tmp_path)
    return str(tmp_path / path.name)


@pytest.fixture(scope="session")
def path_iter_jorek(tmp_path_factory) -> str:
    """Fixture to provide the path to a sample JOREK IMAS dataset."""
    path = Path(iter_jorek())
    tmp_path = tmp_path_factory.mktemp("cherab-imas-data")
    shutil.move(path, tmp_path)
    return str(tmp_path / path.name)



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for loading plasma equilibrium and magnetic field from the equilibrium IDS."""

from typing import Literal, overload

import numpy as np
from raysect.core.math.function.float.function1d import Interpolator1DArray
from raysect.core.math.function.float.function2d import Interpolator2DArray
from raysect.core.math.function.vector3d.function2d import FloatToVector3DFunction2D, Function2D

from cherab.imas.ids.equilibrium.load_equilibrium import Equilibrium2DData
from cherab.tools.equilibrium import EFITEquilibrium
from imas import DBEntry

from ..ids.common import get_ids_time_slice
from ..ids.equilibrium import load_equilibrium_data, load_magnetic_field_data

__all__ = ["load_equilibrium", "load_magnetic_field"]


@overload
def load_equilibrium(
    *args,
    time: float = 0,
    occurrence: int = 0,
    time_threshold: float = np.inf,
    with_psi_interpolator: Literal[False] = False,
    **kwargs,
) -> EFITEquilibrium: ...


@overload
def load_equilibrium(
    *args,
    time: float = 0,
    occurrence: int = 0,
    time_threshold: float = np.inf,
    with_psi_interpolator: Literal[True],
    **kwargs,
) -> tuple[EFITEquilibrium, Interpolator1DArray | None]: ...


def load_equilibrium(
    *args,
    time: float = 0,
    occurrence: int = 0,
    time_threshold: float = np.inf,
    with_psi_interpolator: bool = False,
    **kwargs,
) -> tuple[EFITEquilibrium, Interpolator1DArray | None] | EFITEquilibrium:
    """Load plasma equilibrium from the equilibrium IDS and create an `EFITEquilibrium` object.

    Parameters
    ----------
    *args
        Arguments passed to `~imas.db_entry.DBEntry`.
    time
        Time for the equilibrium, by default 0.
    occurrence
        Occurrence index of the ``equilibrium`` IDS, by default 0.
    time_threshold
        Maximum allowed difference between the requested time and the nearest available
        time, by default `numpy.inf`.
    with_psi_interpolator
        If True, returns the ``psi_norm(rho_tor_norm)`` interpolator; otherwise, returns only the
        equilibrium object.
    **kwargs
        Keyword arguments passed to `~imas.db_entry.DBEntry`.

    Returns
    -------
    equilibrium : `~cherab.tools.equilibrium.efit.EFITEquilibrium`
        The plasma equilibrium object.
    psi_interpolator : `~raysect.core.math.function.float.function1d.interpolate.Interpolator1DArray` | None
        If ``with_psi_interpolator`` is True and ``rho_tor_norm`` is available, returns the
        ``psi_norm(rho_tor_norm)`` interpolator.
        If rho_tor_norm is not available, returns None.
        Otherwise, returns only the equilibrium object.
    """
    with DBEntry(*args, **kwargs) as entry:
        equilibrium_ids = get_ids_time_slice(
            entry, "equilibrium", time=time, occurrence=occurrence, time_threshold=time_threshold
        )

    eq_data = load_equilibrium_data(equilibrium_ids)

    cocos_11to3(eq_data)

    eq_data.psi_norm[0] = min(0, eq_data.psi_norm[0])
    eq_data.psi_norm[-1] = max(1.0, eq_data.psi_norm[-1])

    f_profile = np.array([eq_data.psi_norm, eq_data.f])
    q_profile = np.array([eq_data.psi_norm, eq_data.q])
    equilibrium = EFITEquilibrium(
        eq_data.r,
        eq_data.z,
        eq_data.psi_grid,
        eq_data.psi_axis,
        eq_data.psi_lcfs,
        eq_data.magnetic_axis,
        eq_data.x_points,
        eq_data.strike_points,
        f_profile,
        q_profile,
        eq_data.b_vacuum_radius,
        eq_data.b_vacuum_magnitude,
        eq_data.lcfs_polygon,
        None,  # Limiter Polygon
        eq_data.time,
    )

    if not with_psi_interpolator:
        return equilibrium

    if eq_data.rho_tor_norm is None:
        return equilibrium, None

    psi_interpolator = Interpolator1DArray(
        eq_data.rho_tor_norm,
        eq_data.psi_norm,
        "cubic",
        "none",
        0,
    )

    return equilibrium, psi_interpolator


def load_magnetic_field(
    *args,
    time: float = 0,
    occurrence: int = 0,
    time_threshold: float = np.inf,
    **kwargs,
) -> Function2D:
    """Load the magnetic field from the equilibrium IDS and return an interpolator function.

    Parameters
    ----------
    *args
        Arguments passed to `~imas.db_entry.DBEntry`.
    time
        Time for the equilibrium, by default 0.
    occurrence
        Occurrence index of the ``equilibrium`` IDS, by default 0.
    time_threshold
        Maximum allowed difference between the requested time and the nearest available
        time, by default `numpy.inf`.
    **kwargs
        Keyword arguments passed to `~imas.db_entry.DBEntry`.

    Returns
    -------
    `~raysect.core.math.function.vector3d.function2d.base.Function2D`
        The magnetic field interpolator.

    Raises
    ------
    RuntimeError
        If the equilibrium IDS does not have a time slice.
    """
    with DBEntry(*args, **kwargs) as entry:
        equilibrium_ids = get_ids_time_slice(
            entry, "equilibrium", time=time, occurrence=occurrence, time_threshold=time_threshold
        )

    if not len(equilibrium_ids.time_slice):
        raise RuntimeError("Equilibrium IDS does not have a time slice.")

    b_data = load_magnetic_field_data(equilibrium_ids.time_slice[0].profiles_2d)

    extra_range_r = b_data.r[-1] - b_data.r[0]
    extra_range_z = b_data.z[-1] - b_data.z[0]

    br = Interpolator2DArray(
        b_data.r,
        b_data.z,
        b_data.b_field_r,
        "cubic",
        "linear",
        extra_range_r,
        extra_range_z,
    )
    btor = Interpolator2DArray(
        b_data.r,
        b_data.z,
        b_data.b_field_phi,
        "cubic",
        "linear",
        extra_range_r,
        extra_range_z,
    )
    bz = Interpolator2DArray(
        b_data.r,
        b_data.z,
        b_data.b_field_z,
        "cubic",
        "linear",
        extra_range_r,
        extra_range_z,
    )
    return FloatToVector3DFunction2D(br, btor, bz)


def cocos_11to3(eq_data: Equilibrium2DData) -> None:
    """Convert from COCOS 11 convention used in IMAS to COCOS 3 convention used in EFIT.

    This function modifies the eq_data in place to convert the coordinates
    and other relevant data from the COCOS 11 convention to the COCOS 3 convention.

    Parameters
    ----------
    eq_data
        `Equilibrium2DData` dataclass instance containing equilibrium data to be converted.
    """
    eq_data.psi_grid = -eq_data.psi_grid / (2.0 * np.pi)
    eq_data.psi_axis = -eq_data.psi_axis / (2.0 * np.pi)
    eq_data.psi_lcfs = -eq_data.psi_lcfs / (2.0 * np.pi)
    eq_data.q = -eq_data.q
    if eq_data.phi is not None:
        eq_data.phi = -eq_data.phi



# cython: language_level=3

# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module defining simple interpolators for data defined on a 2D structured grid."""
import numpy as np

from raysect.core.math.vector cimport new_vector3d
from raysect.core.math.cython.utility cimport find_index

cimport cython

__all__ = ["StructGridFunction2D", "StructGridVectorFunction2D"]


cdef class StructGridFunction2D(Function2D):
    """Simple interpolator for the data defined on the 2D structured grid.

    Find the cell containing the point (x, y).
    Return the data value for this cell or the `fill_value` if the points lies outside the grid.

    Parameters
    ----------
    x : (L,) array_like
        The corners of the quadrilateral cells along x axis.
    y : (M,) array_like
        The corners of the quadrilateral cells along y axis.
    grid_data : (L-1, M-1) ndarray
        Array containing data in the grid cells.
    fill_value : float, optional
        Value returned outside the gird, by default 0.
    """

    def __init__(
        self,
        object x not None,
        object y not None,
        np.ndarray grid_data not None,
        double fill_value=0,
    ):

        self._x = np.array(x, dtype=np.float64)
        self._y = np.array(y, dtype=np.float64)

        if self._x.ndim != 1:
            raise ValueError("Argument 'x' must be 1D array.")
        if self._y.ndim != 1:
            raise ValueError("Argument 'y' must be 1D array.")

        if self._x.size < 2:
            raise ValueError("Array 'x' must have at least 2 elements.")
        if self._y.size < 2:
            raise ValueError("Array 'y' must have at least 2 elements.")

        # NOTE: Attention!!! Do not copy grid_data!
        # Attribute self._grid_data must point to the original data array,
        # so as not to re-initialize the interpolator if the user changes data values.

        # populate internal attributes
        if grid_data.shape[0] != self._x.size - 1 or grid_data.shape[1] != self._y.size - 1:
            raise ValueError(
                "The shape of the grid_data array does not match the shape of the grid."
            )

        self._grid_data = grid_data
        self._fill_value = fill_value

        self._x_mv = self._x
        self._y_mv = self._y
        self._grid_data_mv = self._grid_data

    def __getstate__(self):
        return self._grid_data, self._fill_value, self._x, self._y

    def __setstate__(self, state):
        self._grid_data, self._fill_value, self._x, self._y = state
        self._x_mv = self._x
        self._y_mv = self._y
        self._grid_data_mv = self._grid_data

    def __reduce__(self):
        return self.__new__, (self.__class__, ), self.__getstate__()

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef double evaluate(self, double x, double y) except? -1e999:

        cdef int i_x = find_index(self._x_mv, x)
        cdef int i_y = find_index(self._y_mv, y)

        if -1 < i_x < self._x_mv.shape[0] and -1 < i_y < self._y_mv.shape[0]:
            return self._grid_data_mv[i_x, i_y]

        return self._fill_value


cdef class StructGridVectorFunction2D(VectorFunction2D):
    """Simple vector interpolator for the data defined on the 2D structured grid.

    Find the cell containing the point (x, y).
    Return the 3D vector value this cell or the `fill_vector` if the points lies outside the grid.

    Parameters
    ----------
    x : (L,) array_like
        The corners of the quadrilateral cells along x axis.
    y : (M,) array_like
        The corners of the quadrilateral cells along y axis.
    grid_vectors : (3, L-1, M-1) ndarray
        Array containing 3D vectors in the grid cells.
    fill_vector : Vector3D
        3D vector returned outside the gird, by default `Vector3D(0, 0, 0)`.
    """

    def __init__(
        self,
        object x not None,
        object y not None,
        np.ndarray grid_vectors not None,
        Vector3D fill_vector=Vector3D(0, 0, 0),
    ):

        self._x = np.array(x, dtype=np.float64)
        self._y = np.array(y, dtype=np.float64)

        if self._x.ndim != 1:
            raise ValueError("Argument 'x' must be 1D array.")
        if self._y.ndim != 1:
            raise ValueError("Argument 'y' must be 1D array.")

        if self._x.size < 2:
            raise ValueError("Array 'x' must have at least 2 elements.")
        if self._y.size < 2:
            raise ValueError("Array 'y' must have at least 2 elements.")

        # NOTE: Attention!!! Do not copy grid_vectors!
        # Attribute self._grid_vectors must point to the original data array,
        # so as not to re-initialize the interpolator if the user changes data values.

        # populate internal attributes
        if (
            grid_vectors.shape[0] != 3
            or grid_vectors.shape[1] != self._x.size - 1
            or grid_vectors.shape[2] != self._y.size - 1
        ):
            raise ValueError(
                "The shape of the grid_vectors array does not match the shape of the grid."
            )

        self._grid_vectors = grid_vectors
        self._fill_vector = fill_vector

        self._x_mv = self._x
        self._y_mv = self._y
        self._grid_vectors_mv = self._grid_vectors

    def __getstate__(self):
        return self._grid_vectors, self._fill_vector, self._x, self._y

    def __setstate__(self, state):
        self._grid_vectors, self._fill_vector, self._x, self._y = state
        self._x_mv = self._x
        self._y_mv = self._y
        self._grid_vectors_mv = self._grid_vectors

    def __reduce__(self):
        return self.__new__, (self.__class__, ), self.__getstate__()

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef Vector3D evaluate(self, double x, double y):

        cdef int i_x = find_index(self._x_mv, x)
        cdef int i_y = find_index(self._y_mv, y)
        cdef double vx, vy, vz

        if -1 < i_x < self._x_mv.shape[0] and -1 < i_y < self._y_mv.shape[0]:

            vx = self._grid_vectors_mv[0, i_x, i_y]
            vy = self._grid_vectors_mv[1, i_x, i_y]
            vz = self._grid_vectors_mv[2, i_x, i_y]

            return new_vector3d(vx, vy, vz)

        return self._fill_vector



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for loading unstructured 3D grids from IMAS grid_ggd IDS structure."""

from enum import IntEnum

import numpy as np

from imas.ids_struct_array import IDSStructArray
from imas.ids_structure import IDSStructure

from ....ggd import UnstructGrid2DExtended

__all__ = ["load_unstruct_grid_2d_extended", "load_unstruct_grid_3d"]


class SPACE(IntEnum):
    """Enumeration for grid spaces."""

    RZ = 0
    FOURIER = 1


class DIMENSION(IntEnum):
    """Enumeration for grid dimensions."""

    VERTEX = 0
    EDGE = 1
    FACE = 2


# Constants for tetrahedral cells and default number of toroidal points
TETRA_IN_CELL = 5
NUM_TOROIDAL = 64


def load_unstruct_grid_2d_extended(
    grid_ggd: IDSStructure, with_subsets: bool = False, num_toroidal: int | None = NUM_TOROIDAL
) -> UnstructGrid2DExtended:
    """Load unstructured 2D grid extended in 3D from the ``grid_ggd`` structure.

    The grid is created by revolving the 2D cross-section around the torus with evenly spaced
    toroidal angles.

    .. todo::
        `with_subsets` is not yet implemented.

    Parameters
    ----------
    grid_ggd
        The ``grid_ggd`` structure.
    with_subsets
        Read grid subset data if True, by default False.
    num_toroidal
        Number of toroidal points.
        If specifying more than 1, the grid will be extended in 3D by repeating the 2D
        cross-section around the torus with evenly spaced toroidal angles.

    Returns
    -------
    `.UnstructGrid2DExtended`
        The unstructured 2D grid extended in 3D.

    Raises
    ------
    ValueError
        If the number of toroidal points is less than 1.
        If the grid is not an unstructured extended 2D grid.
    """
    # Validate num_toroidal
    num_toroidal = num_toroidal or NUM_TOROIDAL
    if num_toroidal < 1:
        raise ValueError("The number of toroidal points must be greater than 0.")

    # Get the R-Z space
    space: IDSStructArray = grid_ggd.space[SPACE.RZ]

    # Check if the grid is 2D
    if len(space.objects_per_dimension) != 3:
        raise ValueError(
            "The load_unstruct_grid_2d_extended() supports only unstructured extended 2D grids."
        )

    grid_name = str(grid_ggd.identifier.name)

    # =========================================
    # Reading vertices (poloidal and toroidal)
    # =========================================
    num_poloidal = len(space.objects_per_dimension[DIMENSION.VERTEX].object)
    num_vert = num_poloidal * num_toroidal
    vertices_rpz = np.empty((num_vert, 3), dtype=float)

    for i_phi, i_pol in np.ndindex(num_toroidal, num_poloidal):
        phi = 2.0 * np.pi * i_phi / num_toroidal
        vertices_rpz[i_pol + i_phi * num_poloidal] = (
            space.objects_per_dimension[DIMENSION.VERTEX].object[i_pol].geometry[0],
            phi,
            space.objects_per_dimension[DIMENSION.VERTEX].object[i_pol].geometry[1],
        )
    # Convert to cartesian coordinates
    vertices = np.empty((num_vert, 3), dtype=float)
    p = vertices_rpz[:, 1]
    vertices[:, 0] = vertices_rpz[:, 0] * np.cos(p)
    vertices[:, 1] = vertices_rpz[:, 0] * np.sin(p)
    vertices[:, 2] = vertices_rpz[:, 2]

    # =========================================
    # Reading cells indices
    # =========================================
    faces: IDSStructArray = space.objects_per_dimension[DIMENSION.FACE].object
    num_faces = len(faces)
    cells = np.zeros((num_faces * num_toroidal, 8), dtype=np.int32)
    i_cell = 0
    for i_phi, i_face in np.ndindex(num_toroidal, num_faces):
        i_phi_next = i_phi + 1 if i_phi + 1 < num_toroidal else 0
        cells[i_cell] = np.hstack(
            [
                (faces[i_face].nodes - 1) + i_phi * num_poloidal,
                (faces[i_face].nodes - 1) + i_phi_next * num_poloidal,
            ]
        )
        i_cell += 1

    grid = UnstructGrid2DExtended(
        vertices, cells, num_faces, num_poloidal, num_toroidal, name=grid_name
    )

    if not with_subsets:
        return grid
    else:
        raise NotImplementedError("Reading grid subsets is not implemented yet.")


def load_unstruct_grid_3d(grid_ggd: IDSStructure, space_index: int = 0, with_subsets: bool = False):
    """Load unstructured 3D grid from the ``grid_ggd`` structure.

    .. warning::
        This function is a placeholder for future implementation.

    Parameters
    ----------
    grid_ggd
        The ``grid_ggd`` structure.
    space_index
        The index of the space to read, by default 0.
    with_subsets
        Read grid subset data if True, by default False.

    Returns
    -------
    `.UnstructGrid3D`
        The unstructured 3D grid.
    """
    raise NotImplementedError("Loading unstructured 3D grids will be implemented in the future.")



"""Progress bar implementation for pooch using rich."""

try:
    from rich.progress import (
        BarColumn,
        DownloadColumn,
        Progress,
        TaskID,
        TextColumn,
        TimeRemainingColumn,
        TransferSpeedColumn,
    )
except ImportError as e:
    raise ImportError(
        "Missing optional dependency 'rich' required for cherab.imas.datasets module. "
        + "Please use pip or conda to install 'rich'."
    ) from e


class PoochRichProgress:
    def __init__(self, filename: str = ""):
        self.progress: Progress = Progress(
            TextColumn("[bold blue]{task.fields[filename]}", justify="right"),
            BarColumn(bar_width=None),
            "[progress.percentage]{task.percentage:>3.1f}%",
            "•",
            DownloadColumn(),
            "•",
            TransferSpeedColumn(),
            "•",
            TimeRemainingColumn(elapsed_when_finished=True),
        )
        self.task_id: TaskID = self.progress.add_task(description="", filename=filename, total=None)

        # Initialize variables
        self._finished_speed: float | None = 0.0

    @property
    def total(self) -> float | None:
        return self.progress.tasks[self.task_id].total

    @total.setter
    def total(self, value: int) -> None:
        self.progress.update(self.task_id, total=value)
        self.progress.start()

    def close(self) -> None:
        # Restore the finished speed before closing because `reset` is called at the end of
        # a download in pooch.
        self.progress.tasks[self.task_id].finished_speed = self._finished_speed
        self.progress.stop_task(self.task_id)
        self.progress.stop()

    def update(self, chunk_size: int) -> None:
        self.progress.update(self.task_id, advance=chunk_size)

    def reset(self) -> None:
        # Store the finished speed before resetting
        self._finished_speed = self.progress.tasks[self.task_id].speed

        # Reset the progress bar
        self.progress.update(self.task_id, completed=0)



from matplotlib import pyplot as plt
from raysect.core.math.function.vector3d import Function2D as Vector3DFunction2D

from cherab.imas.plasma.equilibrium import load_equilibrium, load_magnetic_field
from cherab.tools.equilibrium import EFITEquilibrium, plot_equilibrium

plt.rcParams["backend"] = "Agg"  # Use non-interactive backend for testing


def test_load_equilibrium(path_iter_jintrac: str):
    """Test loading of equilibrium data from an IMAS file."""
    equilibrium = load_equilibrium(path_iter_jintrac, "r")

    # Test that equilibrium object is returned
    assert isinstance(equilibrium, EFITEquilibrium)

    # Test plotting function
    plot_equilibrium(equilibrium)


def test_load_magnetic_field(path_iter_jintrac: str):
    """Test loading of magnetic field data from an IMAS file."""
    magnetic_field = load_magnetic_field(path_iter_jintrac, "r")

    # Test that Vector3DFunction2D object is returned
    assert isinstance(magnetic_field, Vector3DFunction2D)

    try:
        magnetic_field(0, 0)
        magnetic_field(6.0, 2.5)
        magnetic_field(4.5, -2.5)
    except Exception as err:
        raise RuntimeError("Error at magnetic function call") from err



#!/usr/bin/env -S pixi run python
"""Edge plasma grid plotting demo for ITER-like scenarios.

This demo reads edge plasma grid and profiles and plots the plasma profiles on the grid.
"""

from pathlib import Path

import imas
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.colors import SymLogNorm
from matplotlib.figure import Figure

from cherab.imas.datasets import iter_jintrac
from cherab.imas.ids.common import get_ids_time_slice
from cherab.imas.ids.common.ggd import load_grid
from cherab.imas.ids.edge_profiles import load_edge_species

plt.ion()

grid_subset_name = "Cells"


def plot_grid_quantity(grid, quantity, title="", logscale=False, symmetric=False) -> Figure:
    """Make a 2D plot of a grid quantity, with a title, optionally on a log scale.

    Parameters
    ----------
    grid
        The grid object.
    quantity
        1D array of the quantity to plot on the grid.
    title
        The title of the plot.
    logscale
        Whether to use a logarithmic scale for the color map.
    symmetric
        Whether to use a symmetric color map around zero.

    Returns
    -------
    The matplotlib Figure object containing the plot.
    """
    ax = grid.plot_mesh(data=quantity)

    if logscale:
        linthresh = np.percentile(np.unique(quantity), 1)
        norm = SymLogNorm(linthresh=float(max(linthresh, 1.0e-10 * quantity.max())), base=10)
        ax.collections[0].set_norm(norm)

    if symmetric:
        vmax = np.abs(quantity.max())
        ax.collections[0].set_clim(-vmax, vmax)
        ax.collections[0].set_cmap("coolwarm")
    else:
        ax.collections[0].set_cmap("gnuplot")

    ax.set_title(title)
    ax.set_xlabel("R [m]")
    ax.set_ylabel("Z [m]")
    fig = ax.get_figure()
    fig.colorbar(ax.collections[0], ax=ax, aspect=40)
    fig.set_size_inches((4, 6))

    return fig


demos_path = Path(__file__).parent
plots_path = demos_path / "plots"
plots_path.mkdir(exist_ok=True)

with imas.DBEntry(iter_jintrac(), "r") as entry:
    ids = get_ids_time_slice(entry, "edge_profiles", time=0)

grid, subsets, subset_id = load_grid(ids.grid_ggd[0], with_subsets=True)

try:
    grid = grid.subset(subsets[grid_subset_name], name=grid_subset_name)
except KeyError:
    grid_subset_name = grid_subset_name.lower()
    grid = grid.subset(subsets[grid_subset_name], name=grid_subset_name)

composition = load_edge_species(ids.ggd[0], grid_subset_index=subset_id[grid_subset_name])

# Plot mesh
ax = grid.plot_mesh()
ax.set_xlabel("R [m]")
ax.set_ylabel("Z [m]")
fig = ax.get_figure()
fig.set_size_inches((4, 6))
fig.savefig(plots_path / "edge_mesh.png", dpi=300)

# Plot electron profiles
fig = plot_grid_quantity(
    grid, composition["electron"]["density"], title="Electron density [m-3]", logscale=True
)
fig.savefig(plots_path / "edge_mesh_ne.png", dpi=200)

fig = plot_grid_quantity(
    grid, composition["electron"]["temperature"], title="Electron temperature [eV]", logscale=True
)
fig.savefig(plots_path / "edge_mesh_te.png", dpi=200)

vpar = composition["electron"]["velocity_parallel"]
if np.any(vpar):
    fig = plot_grid_quantity(grid, vpar, title="Electron parallel velocity [m/s]", symmetric=True)
    fig.savefig(plots_path / "edge_mesh_electron_vpar.png", dpi=200)

# Plot species profiles
for stype in ("ion", "molecule"):
    for species_id, profiles in composition[stype].items():
        d = {first: second for first, second in species_id}
        charge = int(round(d["z"]))
        if stype == "ion":
            element = d["element"]
            name = d["name"] if len(d["name"]) else f"{element.symbol}{charge}"
        else:
            element = d["elements"][0]
            name = (
                d["name"]
                if len(d["name"])
                else "{}{}".format("".join([el.symbol for el in d["elements"]]), charge)
            )

        fig = plot_grid_quantity(
            grid, profiles["density"], title=f"{name} density [m-3]", logscale=True
        )
        fig.savefig(plots_path / f"edge_mesh_{name}_density.png", dpi=200)

        if element.atomic_number == 1:
            temperature = profiles["temperature"]
            if np.any(temperature):
                fig = plot_grid_quantity(
                    grid, profiles["temperature"], title=f"{name} temperature [eV]", logscale=True
                )
                fig.savefig(
                    plots_path / f"edge_mesh_{name}_temperature.png",
                    dpi=200,
                )

            if charge:
                vpar = profiles["velocity_parallel"]
                if np.any(vpar):
                    fig = plot_grid_quantity(
                        grid, vpar, title=f"{name} parallel velocity [m/s]", symmetric=True
                    )
                    fig.savefig(
                        plots_path / f"edge_mesh_{name}_vpar.png",
                        dpi=200,
                    )
            else:
                vrad = profiles["velocity_radial"]
                if np.any(vrad):
                    fig = plot_grid_quantity(
                        grid, vrad, title=f"{name} radial velocity [m/s]", symmetric=True
                    )
                    fig.savefig(
                        plots_path / f"edge_mesh_{name}_vrad.png",
                        dpi=200,
                    )
                vpol = profiles["velocity_poloidal"]
                if np.any(vpol):
                    fig = plot_grid_quantity(
                        grid, vrad, title=f"{name} poloidal velocity [m/s]", symmetric=True
                    )
                    fig.savefig(
                        plots_path / f"edge_mesh_{name}_vpol.png",
                        dpi=200,
                    )
                vtor = profiles["velocity_phi"]
                if np.any(vtor):
                    fig = plot_grid_quantity(
                        grid, vtor, title=f"{name} toroidal velocity [m/s]", symmetric=True
                    )
                    fig.savefig(
                        plots_path / f"edge_mesh_{name}_vtor.png",
                        dpi=200,
                    )
